from typing import Any, Literal

from pydantic import Field

from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum
from dhenara.ai.types.shared.base import BaseModel


class CredentialFieldConfig(BaseModel):
    """Configuration for a credential field"""

    field_name: str = Field(..., description="Name of the credential field")
    is_json_field: bool = Field(False, description="Whether the field contains JSON data")  # noqa: FBT003
    error_msg: str = Field(..., description="Error message for validation failures")


class CredentialOutputMapping(BaseModel):
    """Mapping configuration for credential output"""

    source: Literal["api_key", "credentials", "config"] = Field(
        ...,
        description="Source of the credential value",
    )
    source_key: str | None = Field(
        None,
        description="Key in the source dictionary (for credentials and config)",
    )
    output_key: str = Field(
        ...,
        description="Key name in the output credentials dictionary",
    )
    default_value: Any | None = Field(
        None,
        description="Default value if the source is not found",
    )


class ProviderCredentialsConfig(BaseModel):
    """Configuration for provider credentials"""

    api_key_required: bool = Field(..., description="Whether an API key is required")
    credentials_required_fields: list[CredentialFieldConfig] = Field(
        default_factory=list,
        description="Required credential fields",
    )
    credentials_optional_fields: list[CredentialFieldConfig] = Field(
        default_factory=list,
        description="Optional credential fields",
    )
    config_required_fields: list[CredentialFieldConfig] = Field(
        default_factory=list,
        description="Required configuration fields",
    )
    config_optional_fields: list[CredentialFieldConfig] = Field(
        default_factory=list,
        description="Optional configuration fields",
    )
    output_mappings: list[CredentialOutputMapping] = Field(
        default_factory=list,
        description="Mapping configuration for credential output",
    )

    def get_credentials_fields_config_with_json(self) -> list[CredentialFieldConfig]:
        all_credentials = self.credentials_required_fields + self.credentials_optional_fields
        return [field_config for field_config in all_credentials if field_config.is_json_field]


PROVIDER_CONFIGS: dict[AIModelAPIProviderEnum, ProviderCredentialsConfig] = {
    AIModelAPIProviderEnum.OPEN_AI: ProviderCredentialsConfig(
        api_key_required=True,
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
        ],
    ),
    AIModelAPIProviderEnum.GOOGLE_AI: ProviderCredentialsConfig(
        api_key_required=True,
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
        ],
    ),
    AIModelAPIProviderEnum.ANTHROPIC: ProviderCredentialsConfig(
        api_key_required=True,
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
        ],
    ),
    AIModelAPIProviderEnum.DEEPSEEK: ProviderCredentialsConfig(
        api_key_required=True,
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
        ],
    ),
    AIModelAPIProviderEnum.GOOGLE_VERTEX_AI: ProviderCredentialsConfig(
        api_key_required=False,
        credentials_required_fields=[
            CredentialFieldConfig(
                field_name="service_account_json",
                is_json_field=True,
                error_msg="Service account JSON must be a valid dictionary with project_id",
            ),
        ],
        config_required_fields=[
            CredentialFieldConfig(
                field_name="project_id",
                error_msg="Project ID must be a non-empty string",
            ),
            CredentialFieldConfig(
                field_name="location",
                error_msg="Location must be a non-empty string",
            ),
        ],
        output_mappings=[
            CredentialOutputMapping(
                source="credentials",
                source_key="service_account_json",
                output_key="service_account_json",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="project_id",
                output_key="project_id",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="location",
                output_key="location",
                default_value="us-central1",
            ),
        ],
    ),
    AIModelAPIProviderEnum.AMAZON_BEDROCK: ProviderCredentialsConfig(
        api_key_required=False,
        credentials_required_fields=[
            CredentialFieldConfig(
                field_name="access_key_id",
                error_msg="Invalid AWS access key ID format. Should start with 'AKIA'",
            ),
            CredentialFieldConfig(
                field_name="secret_access_key",
                error_msg="Invalid AWS secret access key format. Should be at least 40 characters long",
            ),
        ],
        config_required_fields=[
            CredentialFieldConfig(
                field_name="region",
                error_msg="Region must be a non-empty string",
            ),
        ],
        output_mappings=[
            CredentialOutputMapping(
                source="credentials",
                source_key="access_key_id",
                output_key="aws_access_key",
            ),
            CredentialOutputMapping(
                source="credentials",
                source_key="secret_access_key",
                output_key="aws_secret_key",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="region",
                output_key="aws_region",
            ),
        ],
    ),
    AIModelAPIProviderEnum.MICROSOFT_OPENAI: ProviderCredentialsConfig(
        api_key_required=True,
        credentials_required_fields=[],
        credentials_optional_fields=[],
        config_required_fields=[
            CredentialFieldConfig(
                field_name="endpoint",
                error_msg="endpoint must be a non-empty string",
            ),
        ],
        config_optional_fields=[
            CredentialFieldConfig(
                field_name="api_version",
                error_msg="api_version must be a non-empty string",
            ),
        ],
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="endpoint",
                output_key="azure_endpoint",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="api_version",
                output_key="api_version",
                default_value="2024-10-21",
            ),
        ],
    ),
    AIModelAPIProviderEnum.MICROSOFT_AZURE_AI: ProviderCredentialsConfig(
        api_key_required=True,
        credentials_required_fields=[],
        credentials_optional_fields=[],
        config_required_fields=[
            CredentialFieldConfig(
                field_name="endpoint",
                error_msg="endpoint must be a non-empty string",
            ),
        ],
        config_optional_fields=[],
        output_mappings=[
            CredentialOutputMapping(
                source="api_key",
                output_key="api_key",
            ),
            CredentialOutputMapping(
                source="config",
                source_key="endpoint",
                output_key="azure_endpoint",
            ),
        ],
    ),
}



class DhenaraError(Exception):
    """Base exception for Dhenara API errors"""

    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        response: dict | None = None,
    ):
        super().__init__(message)
        self.status_code = status_code
        self.response = response


class DhenaraAPIError(DhenaraError):
    """Raised when the API returns an error response"""

    pass


class DhenaraConnectionError(DhenaraError):
    """Raised when there's a network connection error"""

    pass


# class ValidationError(APIError):
#    """Raised when the request data is invalid."""
#    pass
#
#
# class AuthenticationError(APIError):
#    """Raised when authentication fails."""
#    pass
#
#
# class PermissionError(APIError):
#    """Raised when the user doesn't have permission."""
#    pass
#
#
# class NotFoundError(APIError):
#    """Raised when a resource is not found."""
#    pass



import logging
from datetime import datetime as datetime_type

from dhenara.ai.config import settings
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    AIModelEndpoint,
    AIModelFunctionalTypeEnum,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseChunk,
    ChatResponseContentItemType,
    ChatResponseGenericContentItem,
    ChatResponseReasoningContentItem,
    ChatResponseTextContentItem,
    ChatResponseToolCallContentItem,
    ChatResponseUsage,
    ExternalApiCallStatus,
    ExternalApiCallStatusEnum,
    ImageResponseUsage,
    StreamingChatResponse,
    UsageCharge,
)
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


class INTStreamingProgress(BaseModel):
    """INTERNAL : Tracks the progress of a streaming response"""

    # total_content: str = ""
    updates_count: int = 0
    start_time: datetime_type
    last_token_time: datetime_type
    is_complete: bool = False
    # Add tracking for Deepseek thinking state, which is embedded in content
    in_thinking_block: bool = False


class StreamingManager:
    """Manages streaming state and constructs final ChatResponse"""

    def __init__(
        self,
        model_endpoint: AIModelEndpoint,
    ):
        self.model_endpoint = model_endpoint

        # Fields required  to create  final ChatResponse
        # self.final_response: ChatResponse | None = None
        self.usage: ChatResponseUsage | None = None
        self.usage_charge: UsageCharge | None = None
        self.choices: list[ChatResponseChoice] = []
        self.response_metadata = AIModelCallResponseMetaData(streaming=True)

        # TODO: cleanup naming
        self.provider_metadata = {}
        self.message_metadata = {}  # Anthropic
        self.persistant_choice_metadata_list = []  # OpenAI

        start_time = datetime_type.now()
        # TODO_FUTURE: Create progress per choices ?
        self.progress = INTStreamingProgress(
            start_time=start_time,
            last_token_time=start_time,
        )

    def update_usage(self, usage: ChatResponseUsage | None = None):
        """Update usgae"""
        if usage:
            self.usage = usage

    def complete(self) -> AIModelCallResponse:
        """Mark streaming as complete and set final stats"""
        self.progress.is_complete = True

        # Calculate duration
        duration = self.progress.last_token_time - self.progress.start_time
        duration_seconds = duration.total_seconds()

        self.response_metadata.duration_seconds = duration_seconds
        self.response_metadata.provider_metadata = self.provider_metadata

        return self.get_final_response()

    def get_final_response(self) -> AIModelCallResponse:
        """Convert streaiming progress to ChatResponse format"""

        chat_response = None

        usage, usage_charge = self.get_streaming_usage_and_charge()

        if self.model_endpoint.ai_model.functional_type == AIModelFunctionalTypeEnum.TEXT_GENERATION:
            chat_response = ChatResponse(
                model=self.model_endpoint.ai_model.model_name,
                provider=self.model_endpoint.ai_model.provider,
                api_provider=self.model_endpoint.api.provider,
                usage=usage,
                usage_charge=usage_charge,
                choices=self.choices,
                metadata=self.response_metadata,
            )
        else:
            logger.fatal("Streaming is only supported for Chat generation models")
            return AIModelCallResponse(
                status=ExternalApiCallStatus(
                    status=ExternalApiCallStatusEnum.INTERNAL_PROCESSING_ERROR,
                    model=self.model_endpoint.ai_model.model_name,
                    api_provider=self.model_endpoint.api.provider,
                    message=(
                        f"Model {self.model_endpoint.ai_model.model_name} not supported for streaming. "
                        "Only Chat models are supported."
                    ),
                    code="error",
                    http_status_code=400,
                ),
            )

        api_call_status = ExternalApiCallStatus(
            status=ExternalApiCallStatusEnum.RESPONSE_RECEIVED_SUCCESS,
            model=self.model_endpoint.ai_model.model_name,
            api_provider=self.model_endpoint.api.provider,
            message="Streaming Completed",
            code="success",
            http_status_code=200,
        )

        return AIModelCallResponse(
            status=api_call_status,
            chat_response=chat_response,
            image_response=None,
        )

    def get_streaming_done_chunk(self):
        return StreamingChatResponse(
            id=None,
            data=ChatResponseChunk(
                model=self.model_endpoint.ai_model.model_name,
                provider=self.model_endpoint.ai_model.provider,
                api_provider=self.model_endpoint.api.provider,
                done=True,
            ),
        )

    def get_streaming_usage_and_charge(
        self,
    ) -> tuple[
        ChatResponseUsage | ImageResponseUsage | None,
        UsageCharge | None,
    ]:
        """Parse the OpenAI response into our standard format"""
        usage_charge = None

        if settings.ENABLE_USAGE_TRACKING or settings.ENABLE_COST_TRACKING:
            if not self.usage:
                logger.fatal("Usage not set before completing streaming.")
                return (None, None)

            if settings.ENABLE_COST_TRACKING:
                usage_charge = self.model_endpoint.calculate_usage_charge(self.usage)

        return (self.usage, usage_charge)

    def update(
        self,
        choice_deltas: list[ChatResponseChoiceDelta],
        response_metadata: dict | None = None,
    ) -> ChatResponseChunk:
        """Update streaming progress with new chunk of deltas"""
        # Update metadata if provided
        if response_metadata:
            self.response_metadata.update(response_metadata)

        # Update last token time
        self.progress.last_token_time = datetime_type.now()

        if settings.ENABLE_STREAMING_CONSOLIDATION and choice_deltas:
            # Initialize choices list if empty
            if not self.choices:
                self.choices = [ChatResponseChoice(index=i, contents=[]) for i in range(len(choice_deltas))]

            # Process each choice delta
            for choice_delta in choice_deltas:
                choice_index = choice_delta.index

                # Ensure we have enough choices initialized
                while len(self.choices) <= choice_index:
                    self.choices.append(ChatResponseChoice(index=len(self.choices), contents=[]))

                choice = self.choices[choice_index]

                # Update choice metadata
                if choice_delta.finish_reason is not None:
                    choice.finish_reason = choice_delta.finish_reason

                if choice_delta.metadata:
                    choice.metadata.update(choice_delta.metadata)

                # Process content deltas if any
                if choice_delta.content_deltas:
                    # Initialize contents list if empty
                    if not choice.contents:
                        choice.contents = []

                    for content_delta in choice_delta.content_deltas:
                        # Find matching content by type and index, or create new
                        matching_content = None

                        # First try to find exact match by type and index
                        for content in choice.contents:
                            if content and content.type == content_delta.type and content.index == content_delta.index:
                                matching_content = content
                                break

                        # If no exact match, try to find by type only
                        if not matching_content:
                            reversed_contents = list(reversed(choice.contents))
                            for content in reversed_contents:
                                if content and content.type == content_delta.type:
                                    matching_content = content
                                    break

                        # If still no match, create new content item
                        if not matching_content:
                            # Create new content based on delta type
                            if content_delta.type == ChatResponseContentItemType.TEXT:
                                matching_content = ChatResponseTextContentItem(
                                    index=content_delta.index,
                                    type=ChatResponseContentItemType.TEXT,
                                    role=content_delta.role,
                                    text="",
                                    metadata=content_delta.metadata,
                                    storage_metadata=content_delta.storage_metadata,
                                    custom_metadata=content_delta.custom_metadata,
                                )
                            elif content_delta.type == ChatResponseContentItemType.REASONING:
                                matching_content = ChatResponseReasoningContentItem(
                                    index=content_delta.index,
                                    type=ChatResponseContentItemType.REASONING,
                                    role=content_delta.role,
                                    thinking_text="",
                                    metadata=content_delta.metadata,
                                    storage_metadata=content_delta.storage_metadata,
                                    custom_metadata=content_delta.custom_metadata,
                                )
                            elif content_delta.type == ChatResponseContentItemType.TOOL_CALL:
                                matching_content = ChatResponseToolCallContentItem(
                                    index=content_delta.index,
                                    type=ChatResponseContentItemType.TOOL_CALL,
                                    role=content_delta.role,
                                    metadata=content_delta.metadata,
                                    storage_metadata=content_delta.storage_metadata,
                                    custom_metadata=content_delta.custom_metadata,
                                )
                            elif content_delta.type == ChatResponseContentItemType.GENERIC:
                                matching_content = ChatResponseGenericContentItem(
                                    index=content_delta.index,
                                    type=ChatResponseContentItemType.GENERIC,
                                    role=content_delta.role,
                                    metadata=content_delta.metadata,
                                    storage_metadata=content_delta.storage_metadata,
                                    custom_metadata=content_delta.custom_metadata,
                                )
                            else:
                                logger.error(f"stream_manager: Unknown content_delta type {content_delta.type}")
                                continue

                            choice.contents.append(matching_content)

                        # Verify type matches
                        if matching_content.type != content_delta.type or matching_content.index != content_delta.index:
                            logger.error(f"stream_manager: Content type mismatch at index {content_delta.index}")
                            continue

                        # Update content based on delta type
                        if content_delta.type == ChatResponseContentItemType.TEXT:
                            delta_text = content_delta.get_text_delta()
                            if delta_text:
                                matching_content.text = (matching_content.text or "") + delta_text

                        elif content_delta.type == ChatResponseContentItemType.REASONING:
                            delta_text = content_delta.get_text_delta()
                            if delta_text:
                                matching_content.thinking_text = (matching_content.thinking_text or "") + delta_text

                        elif content_delta.type in (
                            ChatResponseContentItemType.TOOL_CALL,
                            ChatResponseContentItemType.GENERIC,
                        ):
                            # Update metadata for tool calls and generic content
                            matching_content.metadata.update(content_delta.metadata)

        # Update token count
        self.progress.updates_count += 1

        # Create and return stream chunk
        return ChatResponseChunk(
            model=self.model_endpoint.ai_model.model_name,
            provider=self.model_endpoint.ai_model.provider,
            api_provider=self.model_endpoint.api.provider,
            usage=self.usage,
            usage_charge=self.usage_charge,
            choice_deltas=choice_deltas,
            metadata=self.response_metadata,
            done=False,
        )



import json
from typing import Any, Generic, TypeVar
from uuid import uuid4

from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel

# Type variable for generic data types
T = TypeVar("T", bound=BaseModel)


class SSEEventType(BaseEnum):
    """Types of Server-Sent Events"""

    TOKEN_STREAM = "token_stream"  # Streaming content chunks
    PUSH = "push"  # Push notifications
    ERROR = "error"  # Error events


class SSEErrorCode(BaseEnum):
    server_error = "server_error"
    external_api_error = "external_api_error"
    client_decode_error = "client_decode_error"


class SSEErrorData(BaseModel):
    """Model for error response data"""

    error_code: SSEErrorCode = Field(
        ...,
        description="Error code identifier",
    )
    message: str = Field(
        ...,
        description="Error message",
    )
    details: dict[str, Any] | None = Field(
        default=None,
        description="Additional error details",
    )


class SSEResponse(BaseModel, Generic[T]):
    """Generic SSE response model supporting different data types"""

    event: SSEEventType = Field(
        ...,
        description="SSE Event type",
    )
    data: T = Field(
        ...,
        description="Event data payload",
    )
    id: str | None = Field(
        default=lambda: None,
        # default_factory=lambda: str(uuid4()),
        description="Unique event identifier",
    )
    retry: int | None = Field(
        default=None,
        description="Retry timeout in milliseconds",
    )

    class Config:
        arbitrary_types_allowed = True

    def set_random_id(self) -> str:
        self.id = str(uuid4())

    def is_error(self) -> bool:
        return self.event == SSEEventType.ERROR

    def to_sse_format(self) -> str:
        """Convert to SSE format string"""
        lines = []

        # Add event type
        lines.append(f"event: {self.event}")

        # Add ID
        lines.append(f"id: {self.id}")

        # Add retry if present
        if self.retry is not None:
            lines.append(f"retry: {self.retry}")

        # Add data
        if isinstance(self.data, BaseModel):
            data_str = self.data.model_dump_json()
        elif isinstance(self.data, (dict, list)):
            data_str = json.dumps(self.data)
        else:
            data_str = str(self.data)

        # Handle multi-line data
        for line in data_str.splitlines():
            lines.append(f"data: {line}")  # noqa: PERF401

        return "\n".join(lines) + "\n\n"

    @classmethod
    def parse_sse(cls, sse_str: str, data_type: T | None = None) -> "SSEResponse[Any]":
        """Parse SSE format string into response object"""
        lines = sse_str.strip().split("\n")
        event_data = {
            "event": None,
            "id": None,
            "retry": None,
            "data": None,
        }

        data_lines = []

        for line in lines:
            if not line.strip():
                continue

            if ":" not in line:
                continue

            field, value = line.split(":", 1)
            field = field.strip()
            value = value.lstrip()

            if field == "data":
                data_lines.append(value)
            elif field in event_data:
                event_data[field] = value

        # Parse event type first
        try:
            event_type = SSEEventType(event_data["event"]) if event_data["event"] else SSEEventType.ERROR
        except ValueError:
            return SSEErrorResponse(
                data=SSEErrorData(
                    error_code=SSEErrorCode.client_decode_error,
                    message=f"Invalid event type: {event_data['event']}",
                ),
            )

        # Join and parse data
        if data_lines:
            try:
                raw_data = json.loads("".join(data_lines))

                # Handle different event types
                if event_type == SSEEventType.ERROR:
                    data = SSEErrorData(
                        error_code=SSEErrorCode(raw_data.get("error_code", SSEErrorCode.server_error)),
                        message=raw_data.get("message", "Unknown error"),
                        details=raw_data.get("details"),
                    )
                else:
                    # For normal events, create T object
                    if data_type is not None:
                        data = data_type(**raw_data)
                    else:
                        data = raw_data

            except json.JSONDecodeError:
                return SSEErrorResponse(
                    data=SSEErrorData(
                        error_code=SSEErrorCode.client_decode_error,
                        message="Invalid JSON in stream data",
                    ),
                )
            except Exception as e:
                return SSEErrorResponse(
                    data=SSEErrorData(
                        error_code=SSEErrorCode.client_decode_error,
                        message=f"Failed to parse stream data: {e!s}",
                    ),
                )
        else:
            return SSEErrorResponse(
                data=SSEErrorData(
                    error_code=SSEErrorCode.client_decode_error,
                    message="No data found in SSE event",
                ),
            )

        # Create response with proper typing
        try:
            if event_type == SSEEventType.ERROR:
                return SSEErrorResponse(
                    data=data,
                    id=event_data["id"],
                    retry=int(event_data["retry"]) if event_data["retry"] else None,
                )
            else:
                return cls(
                    event=event_type,
                    data=data,
                    id=event_data["id"],
                    retry=int(event_data["retry"]) if event_data["retry"] else None,
                )
        except Exception as e:
            return SSEErrorResponse(
                data=SSEErrorData(
                    error_code=SSEErrorCode.client_decode_error,
                    message=f"Failed to create SSE response: {e!s}",
                ),
            )


class SSEErrorResponse(SSEResponse[SSEErrorData]):
    """Specialized SSE response for chat streaming"""

    event: SSEEventType = SSEEventType.ERROR
    data: SSEErrorData



from pydantic import Field

from dhenara.ai.types.genai.dhenara import ChatResponse, ImageResponse
from dhenara.ai.types.genai.dhenara.request import Prompt, PromptConfig, PromptMessageRoleEnum
from dhenara.ai.types.shared.base import BaseModel
from dhenara.ai.types.shared.file import GenericFile


class ConversationNode(BaseModel):
    """Represents a single turn in a conversation."""

    user_query: str
    input_files: list[GenericFile] = Field(default_factory=list)
    response: ChatResponse | ImageResponse | None = None
    timestamp: str | None = None

    def get_prompt(
        self,
        max_words_query=None,
        max_words_file=None,
    ) -> Prompt:
        return Prompt(
            role=PromptMessageRoleEnum.USER,
            text=self.user_query,
            files=self.input_files,
            config=PromptConfig(
                max_words_text=max_words_query,
                max_words_file=max_words_file,
            ),
        )

    def get_context(
        self,
        max_words_query=None,
        max_words_file=None,
        max_words_response=None,
    ) -> list[Prompt]:
        question_prompt = Prompt(
            role=PromptMessageRoleEnum.USER,
            text=self.user_query,
            files=self.input_files,
            config=PromptConfig(
                max_words_text=max_words_query,
                max_words_file=max_words_file,
            ),
        )
        response_prompt = self.response.to_prompt(
            max_words_text=max_words_response,
        )

        return [question_prompt, response_prompt]



from abc import ABC, abstractmethod
from typing import Any

from dhenara.ai.types.genai.ai_model import AIModelEndpoint
from dhenara.ai.types.genai.dhenara.request import (
    FormattedPrompt,
    FunctionDefinition,
    FunctionParameter,
    FunctionParameters,
    Prompt,
    PromptMessageRoleEnum,
    StructuredOutputConfig,
    SystemInstruction,
    ToolChoice,
    ToolDefinition,
)
from dhenara.ai.types.shared.file import GenericFile


class BaseFormatter(ABC):
    """
    Formatter for converting Dhenara types to provider-specific formats and vice versa.
    This decouples the conversion logic from the data classes themselves.
    """

    role_map = None

    @classmethod
    def format_prompt(
        cls,
        prompt: str | dict | Prompt,
        model_endpoint: AIModelEndpoint | None = None,
        **kwargs,
    ) -> FormattedPrompt:
        # First convert a prompt to Dhenara Prompt format
        if isinstance(prompt, str):
            # Formatted Prompt
            formatted_prompt = FormattedPrompt(
                role=PromptMessageRoleEnum.USER,
                text=prompt,
            )
            files = []
            max_words_file = None
        else:
            if isinstance(prompt, dict):
                pyd_prompt = Prompt(**prompt)
            elif isinstance(prompt, Prompt):
                pyd_prompt = prompt
            else:
                raise ValueError(f"format_prompt: unknown prompt type {type(prompt)}. prompt={prompt}")

            files = pyd_prompt.files
            max_words_text = pyd_prompt.config.max_words_text if pyd_prompt.config else None
            max_words_file = pyd_prompt.config.max_words_file if pyd_prompt.config else None

            role = pyd_prompt.role
            formatted_text = pyd_prompt.get_formatted_text(
                max_words=max_words_text,
                **kwargs,
            )
            # Formatted Prompt
            formatted_prompt = FormattedPrompt(
                role=role,
                text=formatted_text,
            )

            # Do files sanity checks
            if (files and not isinstance(files, list)) or not all(isinstance(f, GenericFile) for f in files):
                raise ValueError(f"Invalid type {type(files)} for files. Should be list of GenericFile")

        # Convert dhenara formated prompt and files to provider format
        return cls.convert_prompt(
            formatted_prompt=formatted_prompt,
            model_endpoint=model_endpoint,
            files=files,
            max_words_file=max_words_file,
        )

    @classmethod
    def format_context(
        cls,
        context: list[str | dict | Prompt],
        model_endpoint: AIModelEndpoint | None = None,
        **kwargs,
    ) -> list[dict[str, Any]]:
        if not context:
            return []

        formatted_context = [
            cls.format_prompt(
                prompt=prompt,
                model_endpoint=model_endpoint,
                **kwargs,
            )
            for prompt in context
        ]
        return formatted_context

    @classmethod
    def join_instructions(
        cls,
        instructions: list[str | SystemInstruction],
        model_endpoint: AIModelEndpoint | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        if not instructions:
            return None

        def _process_single_instruction(instr):
            if isinstance(instr, str):
                formatted = instr
            else:
                if isinstance(instr, SystemInstruction):
                    _pyd_instr = instr
                elif isinstance(instr, dict):
                    _pyd_instr = SystemInstruction(**instr)
                else:
                    raise ValueError(f"Illegal instruction type {type(instr)}")

                formatted = _pyd_instr.get_formatted_text(**kwargs)
            return formatted

        formatted_instructions = [_process_single_instruction(instr, **kwargs) for instr in instructions]

        joined = " ".join(formatted_instructions)

        return joined

    @classmethod
    def format_instructions(
        cls,
        instructions: list[str | dict | Prompt],
        model_endpoint: AIModelEndpoint | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        # Convert instructions to  Dhenara Promot
        joined_instructions = cls.join_instructions(instructions, **kwargs)

        # Formatted Prompt
        formatted_prompt = FormattedPrompt(
            role=PromptMessageRoleEnum.SYSTEM,
            text=joined_instructions,
        )

        # Convert Dhenara prompt to provider prompt
        # NOTE: Not using `convert_prompt` as some models ( Eg: Beta models) need special handling for instruction
        return cls.convert_instruction_prompt(
            formatted_prompt=formatted_prompt,
            model_endpoint=model_endpoint,
        )

    @classmethod
    def format_tools(
        cls,
        tools: list[ToolDefinition] | None,
        model_endpoint: AIModelEndpoint | None = None,
    ):
        if tools:
            return [
                cls.format_tool(
                    tool=tool,
                    model_endpoint=model_endpoint,
                )
                for tool in tools
            ]

    @classmethod
    def format_tool(
        cls,
        tool: ToolDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        return cls.convert_tool(
            tool=tool,
            model_endpoint=model_endpoint,
        )

    @classmethod
    def format_tool_choice(
        cls,
        tool_choice: ToolChoice,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        return cls.convert_tool_choice(
            tool_choice=tool_choice,
            model_endpoint=model_endpoint,
        )

    @classmethod
    def format_structured_output(
        cls,
        structured_output: StructuredOutputConfig,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        return cls.convert_structured_output(
            structured_output=structured_output,
            model_endpoint=model_endpoint,
        )

    # Abstract methods
    @classmethod
    @abstractmethod
    def convert_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
        files: list[GenericFile] | None = None,
        max_words_file: int | None = None,
    ) -> dict[str, Any]:
        """Convert prompt to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_instruction_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert instruction prompt to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_files_to_provider_content(
        cls,
        files: list[GenericFile],
        model_endpoint: AIModelEndpoint | None = None,
        max_words: int | None = None,
    ) -> list[dict[str, Any]]:
        """Convert File content to provider format"""
        pass

    # -------------------------------------------------------------------------
    # Tools & Structured output
    @classmethod
    @abstractmethod
    def convert_function_parameter(
        cls,
        param: FunctionParameter,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert a single FunctionParameter to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_function_parameters(
        cls,
        params: FunctionParameters,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameters to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_function_definition(
        cls,
        func_def: FunctionDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionDefinition to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_tool(
        cls,
        tool: ToolDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolDefinition to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_tool_choice(
        cls,
        tool_choice: ToolChoice,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolChoice to provider format"""
        pass

    @classmethod
    @abstractmethod
    def convert_structured_output(
        cls,
        structured_output: StructuredOutputConfig,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert StructuredOutputConfig to provider format"""
        pass



import logging
from collections.abc import AsyncGenerator, Generator
from typing import Union

from pydantic import Field

from dhenara.ai.types.genai.dhenara.response import (
    ChatResponse,
    ExternalApiCallStatus,
    ImageResponse,
    StreamingChatResponse,
)
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


class AIModelCallResponse(BaseModel):
    """
    Response model for AI model calls including both streaming and non-streaming responses.

    Attributes:
        status: Current status of the API call
        chat_response: Response for non-streaming chat API calls
        stream_generator: Async generator for streaming chat responses
        image_response: Response for image generation API calls
    """

    status: ExternalApiCallStatus | None = Field(
        default=None,
        description="API Call status. Will be None with stream generators",
    )
    chat_response: ChatResponse | None = Field(
        default=None,
        description="Response for Non-streaming chat creation API calls",
    )
    async_stream_generator: (
        AsyncGenerator[
            tuple[
                StreamingChatResponse | None,
                Union["AIModelCallResponse", None],
            ]
        ]
        | None
    ) = Field(
        default=None,
        description="""Response for streaming chat creation API calls.
        This will be an async generator that generates the response stream, and on the last chunk
        along with the full response on the last chunk""",
    )
    sync_stream_generator: (
        Generator[
            tuple[
                StreamingChatResponse | None,
                Union["AIModelCallResponse", None],
            ]
        ]
        | None
    ) = Field(
        default=None,
        description="""Sync response for streaming chat creation API calls.
        This will be an async generator that generates the response stream, and on the last chunk
        along with the full response on the last chunk""",
    )
    image_response: ImageResponse | None = Field(
        default=None,
        description="Response for Non-streaming chat creation API calls",
    )

    @property
    def full_response(self) -> ChatResponse | ImageResponse | None:
        """
        Get the full response from either chat or image response.

        Returns:
            ChatResponse | ImageResponse | None: The complete response object
        """
        return self.chat_response if self.chat_response else self.image_response

    @property
    def stream_generator(self) -> AsyncGenerator | Generator | None:
        return self.async_stream_generator if self.async_stream_generator else self.sync_stream_generator

    def preview_dict(self):
        """
        Returns a preview version of the response excluding the choices
        """
        if self.full_response is None:
            return None

        return self.full_response.preview_dict()



import logging

from google.genai.types import (
    GenerateContentConfig,
    GenerateContentResponse,
    Part,
    SafetySetting,
    Tool,
    ToolConfig,
)

# Copyright 2024-2025 Dhenara Inc. All rights reserved.
from dhenara.ai.providers.google import GoogleAIClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseContentItem,
    ChatResponseContentItemDelta,
    ChatResponseGenericContentItem,
    ChatResponseGenericContentItemDelta,
    ChatResponseStructuredOutput,
    ChatResponseStructuredOutputContentItem,
    ChatResponseTextContentItem,
    ChatResponseTextContentItemDelta,
    ChatResponseToolCall,
    ChatResponseToolCallContentItem,
    ChatResponseUsage,
    StreamingChatResponse,
)
from dhenara.ai.types.shared.api import SSEErrorResponse

logger = logging.getLogger(__name__)

logger = logging.getLogger(__name__)


models_not_supporting_system_instructions = ["gemini-1.0-pro"]


# -----------------------------------------------------------------------------
class GoogleAIChat(GoogleAIClientBase):
    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: dict | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("inputs must be validated with `self.validate_inputs()` before api calls")

        generate_config_args = self.get_default_generate_config_args()
        generate_config = GenerateContentConfig(**generate_config_args)

        # Process instructions

        if instructions:
            if not (isinstance(instructions, dict) and "parts" in instructions.keys()):
                raise ValueError(
                    f"Invalid Instructions format. "
                    f"Instructions should be processed and passed in prompt format. Value is {instructions} "
                )

            # Some models don't support system instructions
            if any(self.model_endpoint.ai_model.model_name.startswith(model) for model in ["gemini-1.0-pro"]):
                instruction_as_prompt = instructions

                if context:
                    context.insert(0, instruction_as_prompt)
                else:
                    context = [instruction_as_prompt]
            else:
                instructions_str = instructions["parts"][0]["text"]
                generate_config.system_instruction = instructions_str

        messages = []

        # Add previous messages and current prompt
        if context:
            messages.extend(context)

        messages.append(prompt)

        # ---  Tools ---
        if self.config.tools:
            # NOTE: Google supports extra tools other than fns, so gather all fns together into function_declarations
            # --  _tools = [tool.to_google_format() for tool in self.config.tools]
            _tools = [
                Tool(
                    **{
                        "function_declarations": [
                            self.formatter.convert_function_definition(  # A bit wiered here
                                func_def=tool.function,
                                model_endpoint=self.model_endpoint,
                            )
                            for tool in self.config.tools
                        ],
                    }
                )
            ]
            generate_config.tools = _tools

        if self.config.tool_choice:
            _tool_config = self.formatter.format_tool_choice(
                tool_choice=self.config.tool_choice,
                model_endpoint=self.model_endpoint,
            )

            generate_config.tool_config = ToolConfig(**_tool_config)

        # --- Structured Output ---
        if self.config.structured_output:
            generate_config.response_mime_type = "application/json"
            generate_config.response_schema = self.formatter.format_structured_output(
                structured_output=self.config.structured_output,
                model_endpoint=self.model_endpoint,
            )

        return {
            "contents": messages,
            "generate_config": generate_config,
        }

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        response = self._client.models.generate_content(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            contents=api_call_params["contents"],
        )
        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        response = await self._client.models.generate_content(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            contents=api_call_params["contents"],
        )
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        stream = self._client.models.generate_content_stream(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            contents=api_call_params["contents"],
        )
        return stream

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        stream = await self._client.models.generate_content_stream(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            contents=api_call_params["contents"],
        )
        return stream

    def get_default_generate_config_args(self) -> dict:
        max_output_tokens, max_reasoning_tokens = self.config.get_max_output_tokens(self.model_endpoint.ai_model)
        safety_settings = [
            SafetySetting(
                category="HARM_CATEGORY_DANGEROUS_CONTENT",
                threshold="BLOCK_ONLY_HIGH",
            )
        ]

        config_params = {
            "candidate_count": 1,
            "safety_settings": safety_settings,
        }

        if max_output_tokens:
            config_params["max_output_tokens"] = max_output_tokens

        return config_params

    def parse_stream_chunk(
        self,
        chunk: GenerateContentResponse,
    ) -> StreamingChatResponse | SSEErrorResponse | None:
        """Handle streaming response with progress tracking and final response"""

        processed_chunks = []

        self.streaming_manager.provider_metadata = None

        # Process content
        if chunk.candidates:
            choice_deltas = []
            for candidate_index, candidate in enumerate(chunk.candidates):
                content_deltas = []
                for part_index, part in enumerate(candidate.content.parts):
                    content_deltas.append(
                        self.process_content_item_delta(
                            index=part_index,
                            role=candidate.content.role,
                            delta=part,
                        )
                    )
                choice_deltas.append(
                    ChatResponseChoiceDelta(
                        index=candidate_index,
                        finish_reason=candidate.finish_reason,
                        stop_sequence=None,
                        content_deltas=content_deltas,
                        metadata={"safety_ratings": candidate.safety_ratings, "": candidate.content},  # Choice metadata
                    )
                )

            response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
            stream_response = StreamingChatResponse(
                id=None,  # No 'id' from google
                data=response_chunk,
            )

            processed_chunks.append(stream_response)

            # Check if this is the final chunk
            is_done = bool(candidate.finish_reason)

            if is_done:
                usage = self._get_usage_from_provider_response(chunk)
                self.streaming_manager.update_usage(usage)

        return processed_chunks

    def _get_usage_from_provider_response(
        self,
        response: GenerateContentResponse,
    ) -> ChatResponseUsage:
        candidates_tokens = response.usage_metadata.candidates_token_count or 0
        thoughts_tokens = (
            (response.usage_metadata.thoughts_token_count or 0)
            if hasattr(response.usage_metadata, "thoughts_token_count")
            else 0
        )
        tool_use_tokens = (
            (response.usage_metadata.tool_use_prompt_token_count or 0)
            if hasattr(response.usage_metadata, "tool_use_prompt_token_count")
            else 0
        )

        completion_tokens = candidates_tokens + thoughts_tokens + tool_use_tokens

        return ChatResponseUsage(
            total_tokens=response.usage_metadata.total_token_count or 0,
            prompt_tokens=response.usage_metadata.prompt_token_count or 0,
            completion_tokens=completion_tokens,
        )

    def parse_response(
        self,
        response: GenerateContentResponse,
    ) -> ChatResponse:
        usage, usage_charge = self.get_usage_and_charge(response)
        return ChatResponse(
            model=response.model_version,
            provider=self.model_endpoint.ai_model.provider,
            api_provider=self.model_endpoint.api.provider,
            usage=usage,
            usage_charge=usage_charge,
            choices=[
                ChatResponseChoice(
                    index=choice_index,
                    finish_reason=candidate.finish_reason,
                    stop_sequence=None,
                    contents=[
                        self.process_content_item(
                            index=part_index,
                            role=candidate.content.role,
                            content_item=part,
                        )
                        for part_index, part in enumerate(candidate.content.parts)
                    ],
                    metadata={},  # Choice metadata
                )
                for choice_index, candidate in enumerate(response.candidates)
            ],
            metadata=AIModelCallResponseMetaData(
                streaming=False,
                duration_seconds=0,  # TODO
                provider_metadata={},
            ),
        )

    def process_content_item(
        self,
        index: int,
        role: str,
        content_item: Part,
    ) -> ChatResponseContentItem:
        if isinstance(content_item, Part):
            if hasattr(content_item, "text") and content_item.text is not None:
                _content = content_item.text

                if self.config.structured_output is not None:
                    structured_output = ChatResponseStructuredOutput.from_model_output(
                        raw_response=_content,
                        config=self.config.structured_output,
                    )
                    return ChatResponseStructuredOutputContentItem(
                        index=index,
                        role=role,
                        structured_output=structured_output,
                    )
                else:
                    return ChatResponseTextContentItem(
                        index=index,
                        role=role,
                        text=_content,
                    )

            elif hasattr(content_item, "function_call") and content_item.function_call is not None:
                content_item_dict = content_item.function_call.model_dump()

                return ChatResponseToolCallContentItem(
                    index=index,
                    role=role,
                    tool_call=ChatResponseToolCall.from_google_format(content_item_dict),
                    metadata={},
                )

            else:
                return ChatResponseGenericContentItem(
                    index=index,
                    role=role,
                    metadata={"part": content_item.model_dump()},
                )
        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=content_item,
                streaming=False,
            )

    # Streaming
    def process_content_item_delta(
        self,
        index: int,
        role: str,
        delta,
    ) -> ChatResponseContentItemDelta:
        if isinstance(delta, Part):
            if hasattr(delta, "text"):
                return ChatResponseTextContentItemDelta(
                    index=index,
                    role=role,
                    text_delta=delta.text,
                )

            # TODO: Tools Not supported in streaming yet
            else:
                return ChatResponseGenericContentItemDelta(
                    index=index,
                    role=role,
                    metadata={"part": delta.model_dump()},
                )

        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=delta,
                streaming=True,
            )



import logging
from typing import Any

from dhenara.ai.providers.base import BaseFormatter
from dhenara.ai.types.genai.ai_model import AIModelEndpoint
from dhenara.ai.types.genai.dhenara.request import (
    FunctionDefinition,
    FunctionParameter,
    FunctionParameters,
    PromptMessageRoleEnum,
    StructuredOutputConfig,
    ToolChoice,
    ToolDefinition,
)
from dhenara.ai.types.genai.dhenara.request.data import FormattedPrompt
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


class AnthropicFormatter(BaseFormatter):
    """
    Formatter for converting Dhenara types to Anthropic-specific formats and vice versa.
    """

    role_map = {
        PromptMessageRoleEnum.USER: "user",
        PromptMessageRoleEnum.ASSISTANT: "assistant",
        PromptMessageRoleEnum.SYSTEM: "system",  # NOTE: Don't care as system instructions are taken care separately
    }

    @classmethod
    def convert_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
        files: list[GenericFile] | None = None,
        max_words_file: int | None = None,
    ) -> dict[str, Any]:
        # Map Dhenara formats to provider format
        file_contents = None
        if files:
            file_contents = cls.convert_files_to_provider_content(
                files=files,
                model_endpoint=model_endpoint,
                max_words=max_words_file,
            )

        if file_contents:
            content = [
                {
                    "type": "text",
                    "text": formatted_prompt.text,
                },
                *file_contents,
            ]
        else:
            # Use Simple text formtat
            content = formatted_prompt.text

        role = cls.role_map.get(formatted_prompt.role)
        return {"role": role, "content": content}

    @classmethod
    def convert_instruction_prompt(
        cls,
        formatted_prompt: FormattedPrompt,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        # There is no native support for `system` prompt. So set role always send as `user` role,
        # so that beta models can send them as prompt.
        # For other models, the text will be send as seperate argument
        role = cls.role_map.get(PromptMessageRoleEnum.USER)

        return {"role": role, "content": formatted_prompt.text}

    @classmethod
    def convert_files_to_provider_content(
        cls,
        files: list[GenericFile],
        model_endpoint: AIModelEndpoint | None = None,
        max_words: int | None = None,
    ) -> list[dict[str, Any]]:
        contents = []

        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = (
                        f"\nFile: {file.get_source_file_name()}  "
                        f"Content: {file.get_processed_file_data(max_words=max_words)}"
                    )
                    contents.append(
                        {
                            "type": "text",
                            "text": text,
                        }
                    )
                elif file_format == FileFormatEnum.IMAGE:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type}")

                    contents.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": mime_type,
                                "data": file.get_processed_file_data_content_only(),
                            },
                        }
                    )
                else:
                    logger.error(f"Unknown file_format {file_format} for file {file.name}")
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return contents

    # -------------------------------------------------------------------------
    # Tools & Structured output
    @classmethod
    def convert_function_parameter(
        cls,
        param: FunctionParameter,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        result = param.model_dump(
            exclude={"required", "allowed_values", "default"},
        )
        if param.allowed_values is not None:
            result["enum"] = param.allowed_values

        return result

    @classmethod
    def convert_function_parameters(
        cls,
        params: FunctionParameters,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert FunctionParameters to Anthropic format"""
        # Create a new dictionary with transformed properties
        result = {
            "type": params.type,
            "properties": {name: cls.convert_function_parameter(param) for name, param in params.properties.items()},
        }

        # Auto-build the required list based on parameters marked as required
        required_params = [name for name, param in params.properties.items() if param.required]

        # Only include required field if there are required parameters
        if required_params:
            result["required"] = required_params
        elif params.required:  # If manually specified required array exists
            result["required"] = params.required

        return result

    @classmethod
    def convert_function_definition(
        cls,
        func_def: FunctionDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """Convert to Anthropic format"""
        return {
            "name": func_def.name,
            "description": func_def.description,
            "input_schema": cls.convert_function_parameters(func_def.parameters),
        }

    @classmethod
    def convert_tool(
        cls,
        tool: ToolDefinition,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolDefinition to Anthropic format"""
        return cls.convert_function_definition(tool.function)

    @classmethod
    def convert_tool_choice(
        cls,
        tool_choice: ToolChoice,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> Any:
        """Convert ToolChoice to Anthropic format"""

        if tool_choice.type is None:
            return None
        elif tool_choice.type == "zero_or_more":
            return {"type": "auto"}
        elif tool_choice.type == "one_or_more":
            return {"type": "any"}
        elif tool_choice.type == "specific":
            return {"type": "tool", "name": tool_choice.specific_tool_name}

    @classmethod
    def convert_structured_output(
        cls,
        structured_output: StructuredOutputConfig,
        model_endpoint: AIModelEndpoint | None = None,
    ) -> dict[str, Any]:
        """
        Convert structured output config to Anthropic tool format.
        Since Anthropic doesn't directly support structured output,
        we create a specialized tool and force the model to use it.
        """
        schema = structured_output.get_schema()
        name = schema.pop("title", None) or "structured_output"
        description = "Generate structured output according to schema"

        # Clean up schema for Anthropic compatibility
        if "properties" in schema:
            for prop in schema["properties"].values():
                # Remove unsupported validations
                for field in ["minimum", "maximum", "exclusiveMinimum", "exclusiveMaximum"]:
                    if field in prop:
                        prop.pop(field)

                # Convert enum to Anthropic format
                if "enum" in prop:
                    prop["enum"] = list(prop["enum"])

        # Create the tool
        tool = {
            "name": name,
            "description": description,
            "input_schema": schema,
        }

        return tool



from dhenara.ai.types.shared.base import BaseEnum, BaseModel


# -----------------------------------------------------------------------------
class FileFormatEnum(BaseEnum):
    TEXT = "text"
    IMAGE = "image"
    AUDIO = "audio"
    VIDEO = "video"
    COMPRESSED = "compressed"
    UNDEFINED = "undefined"


# -----------------------------------------------------------------------------
class FileContentFormatEnum(BaseEnum):
    TEXT = "text"
    BINARY = "binary"
    RAW = "raw"
    UNDEFINED = "undefined"


# -----------------------------------------------------------------------------
class FileProcessingStatusEnum(BaseEnum):
    SUCCESS = "success"
    FAIL = "fail"
    UNDEFINED = "undefined"


# -----------------------------------------------------------------------------
class FileMetadata(BaseModel):
    file_name: str
    file_size: int | float
    mime_type: str | None  # TODO: Enforce after fixing current mime gussess
    isinstance_of_dj_fileclass: bool | None = None  # TODO:Remove this internal flag


# -----------------------------------------------------------------------------
class FileContentData(BaseModel):
    name: str
    metadata: FileMetadata | None
    file_format: FileFormatEnum
    content_format: FileContentFormatEnum
    content: str | None
    contents: list["FileContentData"] | None
    folder_structure: str | None  # Represent the folder structure in case of a compressed file
    processing_status: FileProcessingStatusEnum
    processing_message: str

    def get_content(self):
        if self.file_format != FileFormatEnum.COMPRESSED:
            return self.content or ""
        elif self.contents:
            all_contents = "".join([content.get_content() for content in self.contents])
            return all_contents
        else:
            return ""

    def get_dict_without_content(self):
        if self.file_format != FileFormatEnum.COMPRESSED:
            return {
                "name": self.name,
                "metadata": self.metadata,
                "file_format": self.file_format.value,
                "content_format": self.content_format.value,
                "folder_structure": self.folder_structure,
                "processing_status": self.processing_status.value,
                "processing_message": self.processing_message,
            }
        elif self.contents:
            all_contents = "".join([content.get_dict_without_content() for content in self.contents])
            return {
                "name": self.name,
                "metadata": self.metadata,
                "file_format": self.file_format.value,
                "content_format": self.content_format.value,
                "folder_structure": self.folder_structure,
                "processing_status": self.processing_status.value,
                "processing_message": self.processing_message,
                "file_contents": all_contents,
            }
        else:
            return ""



from typing import Any

from pydantic import Field, model_validator

from dhenara.ai.types.shared.base import BaseModel
from dhenara.ai.types.shared.file import GenericFile

from ._content import Content
from ._role import PromptMessageRoleEnum
from ._text_template import TextTemplate


class PromptText(BaseModel):
    content: Content | None = Field(
        default=None,
        description="Prompt Content",
    )
    template: TextTemplate | None = Field(
        default=None,
        description="Text template with optional {placeholders} for string formatting",
    )

    @model_validator(mode="after")
    def validate_all(self) -> "PromptText":
        if not (self.content or self.template):
            raise ValueError("Content or Template is required for prompt")
        if self.content and self.template:
            raise ValueError("Only one of Content or Template is allowed")
        return self

    def format(self, **kwargs) -> str:
        if self.content:
            return self.content.get_content()
        else:
            return self.template.format(**kwargs)


class PromptConfig(BaseModel):
    max_words_text: int | None = Field(
        default=None,
        description="Max number of words in text. Leave as None for not limiting.",
    )
    max_words_file: int | None = Field(
        default=None,
        description="Max number of words in a single file. Leave as None for not limiting.",
    )


class BaseTextPrompt(BaseModel):
    text: str | PromptText = Field(
        ...,
        description="Prompt Text.",
    )
    variables: dict[str, Any] = Field(
        default_factory=dict,
        description="Variable name s and values for template resolution in prompt.",
        example={"style": "modern", "name": "Annie"},
    )

    def get_formatted_text(
        self,
        max_words: int | None = None,
        **kwargs,
    ) -> str:
        """Format the prompt as a generic dictionary"""
        if isinstance(self.text, PromptText):
            var_dict = self.variables.copy()
            var_dict.update(**kwargs)
            formatted_text = self.text.format(**var_dict)
        elif isinstance(self.text, str):
            formatted_text = self.text
        else:
            raise ValueError(f"get_formatted_text: unknown prompt.text type {type(self.text)}")

        if max_words:
            words = formatted_text.split()
            formatted_text = " ".join(words[:max_words])

        return formatted_text


class Prompt(BaseTextPrompt):
    role: PromptMessageRoleEnum = Field(
        ...,
        description="Role",
    )
    files: list[GenericFile] = Field(
        default_factory=list,
        description="Files",
    )
    config: PromptConfig | None = Field(
        default=None,
        description="Prompt Config.",
    )

    @classmethod
    def with_text(
        cls,
        text: str,
        variables: dict | None = None,
        disable_checks: bool = False,
    ):
        return cls(
            role=PromptMessageRoleEnum.USER,
            text=PromptText(
                content=None,
                template=TextTemplate(
                    text=text,
                    variables=variables or {},
                    disable_checks=disable_checks,
                ),
            ),
        )

    @classmethod
    def with_dad_text(
        cls,
        text: str,
        variables: dict | None = None,
        disable_checks: bool = True,
    ):
        return cls(
            role=PromptMessageRoleEnum.USER,
            text=PromptText(
                content=None,
                template=TextTemplate(
                    text=text,
                    variables=variables or {},
                    disable_checks=disable_checks,
                ),
            ),
        )


class FormattedPrompt(BaseModel):
    role: PromptMessageRoleEnum
    text: str


class SystemInstruction(BaseTextPrompt):
    pass



from dhenara.ai.types.shared.base import BaseEnum


class AIModelFunctionalTypeEnum(BaseEnum):
    TEXT_GENERATION = "text_generation"
    IMAGE_GENERATION = "image_generation"
    VIDEO_GENERATION = "video_generation"
    TEXT_TO_SPEECH = "text_to_speech"
    SPEECH_TO_TEXT = "speech_to_text"


class AIModelProviderEnum(BaseEnum):
    CUSTOM = "custom"
    OPEN_AI = "open_ai"
    GOOGLE_AI = "google_ai"
    ANTHROPIC = "anthropic"
    DEEPSEEK = "deepseek"
    META = "meta"
    COHERE = "cohere"


class AIModelAPIProviderEnum(BaseEnum):
    OPEN_AI = "openai"
    GOOGLE_AI = "google_gemini_api"
    ANTHROPIC = "anthropic"
    GOOGLE_VERTEX_AI = "google_vertex_ai"
    MICROSOFT_OPENAI = "microsoft_openai"
    MICROSOFT_AZURE_AI = "microsoft_azure_ai"
    AMAZON_BEDROCK = "amazon_bedrock"
    DEEPSEEK = "deepseek"



import logging

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.providers.shared import APIProviderSharedFns
from dhenara.ai.types.genai.ai_model import AIModelAPIProviderEnum

from .formatter import AnthropicFormatter

logger = logging.getLogger(__name__)


class AnthropicClientBase(AIModelProviderClientBase):
    """Base class for all Anthropic Clients"""

    formatter = AnthropicFormatter

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""
        if api.provider == AIModelAPIProviderEnum.ANTHROPIC:
            params = {
                "api_key": api.api_key,
                **self._get_client_http_params(api),
            }
            return "anthropic", params

        elif api.provider == AIModelAPIProviderEnum.GOOGLE_VERTEX_AI:
            client_params = APIProviderSharedFns.get_vertex_ai_credentials(api)
            params = {
                "credentials": client_params["credentials"],
                "project_id": client_params["project_id"],
                "region": client_params["location"],
                **self._get_client_http_params(api),
            }
            return "vertex_ai", params

        elif api.provider == AIModelAPIProviderEnum.AMAZON_BEDROCK:
            client_params = api.get_provider_credentials()
            params = {
                "aws_access_key": client_params["aws_access_key"],
                "aws_secret_key": client_params["aws_secret_key"],
                "aws_region": client_params.get("aws_region", "us-east-1"),
                **self._get_client_http_params(api),
            }
            return "bedrock", params

        error_msg = f"Unsupported API provider {api.provider} for Anthropic"
        logger.error(error_msg)
        raise ValueError(error_msg)

    def _setup_client_sync(self) -> Anthropic | AnthropicBedrock | AnthropicVertex:
        """Get the appropriate sync Anthropic client based on the provider"""
        client_type, params = self._get_client_params(self.model_endpoint.api)

        if client_type == "anthropic":
            return Anthropic(**params)
        elif client_type == "vertex_ai":
            return AnthropicVertex(**params)
        else:  # bedrock
            return AnthropicBedrock(**params)

    async def _setup_client_async(self) -> AsyncAnthropic | AsyncAnthropicBedrock | AsyncAnthropicVertex:
        """Get the appropriate async Anthropic client based on the provider"""
        client_type, params = self._get_client_params(self.model_endpoint.api)

        if client_type == "anthropic":
            return AsyncAnthropic(**params)
        elif client_type == "vertex_ai":
            return AsyncAnthropicVertex(**params)
        else:  # bedrock
            return AsyncAnthropicBedrock(**params)



import asyncio
import logging
from collections.abc import AsyncGenerator, Generator

from dhenara.ai.config import settings
from dhenara.ai.providers.base import StreamingManager
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    AIModelFunctionalTypeEnum,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseTextContentItem,
    ChatResponseTextContentItemDelta,
    ChatResponseUsage,
    ExternalApiCallStatus,
    ExternalApiCallStatusEnum,
    ImageResponseUsage,
    StreamingChatResponse,
    UsageCharge,
)
from dhenara.ai.types.genai.ai_model import AIModelEndpoint
from dhenara.ai.types.shared.api import SSEErrorCode, SSEErrorData, SSEErrorResponse

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class DummyAIModelResponseFns:
    def __init__(self, streaming_manager: StreamingManager):
        self.streaming_manager = streaming_manager

    def _create_base_response(self, ai_model_ep: AIModelEndpoint, text: str) -> AIModelCallResponse:
        """Create a base response with common elements"""
        api_call_status = ExternalApiCallStatus(
            status=ExternalApiCallStatusEnum.RESPONSE_RECEIVED_SUCCESS,
            api_provider="dummy_ai_provider",
            model="dummy_model",
            message="Output generated",
            code="success",
            http_status_code=200,
        )

        usage, usage_charge = self.get_dummy_usage_and_charge(ai_model_ep)

        parsed_response = ChatResponse(
            model="dummy_model",
            provider=ai_model_ep.ai_model.provider,
            api_provider=ai_model_ep.api.provider,
            usage=usage,
            usage_charge=usage_charge,
            choices=[
                ChatResponseChoice(
                    index=0,
                    contents=[
                        ChatResponseTextContentItem(
                            role="assistant",
                            text=text,
                        )
                    ],
                )
            ],
            metadata=AIModelCallResponseMetaData(
                streaming=False,
                duration_seconds=0,
                provider_metadata={
                    "id": "test_msg_id",
                    "object": "test_object",
                    "created": "",
                    "system_fingerprint": "",
                },
            ),
        )

        return AIModelCallResponse(
            status=api_call_status,
            chat_response=parsed_response,
        )

    def get_dummy_ai_model_response_sync(
        self,
        ai_model_ep: AIModelEndpoint,
        streaming: bool = False,
    ) -> AIModelCallResponse:
        if streaming:
            _stream_generator = self.handle_streaming_response_sync(
                self.sync_stream_generator(ai_model_ep),
                model_endpoint=ai_model_ep,
            )
            return AIModelCallResponse(sync_stream_generator=_stream_generator)
        return self._create_base_response(
            ai_model_ep,
            f"This is a test mode output. {self._model_specific_message(ai_model_ep)}",
        )

    async def get_dummy_ai_model_response_async(
        self,
        ai_model_ep: AIModelEndpoint,
        streaming: bool = False,
    ) -> AIModelCallResponse:
        if streaming:
            _stream_generator = self.handle_streaming_response_async(
                self.async_stream_generator(ai_model_ep),
                model_endpoint=ai_model_ep,
            )
            return AIModelCallResponse(async_stream_generator=_stream_generator)
        return self._create_base_response(
            ai_model_ep,
            f"This is a test mode output. {self._model_specific_message(ai_model_ep)}",
        )

    def get_dummy_usage_and_charge(
        self,
        ai_model_ep: AIModelEndpoint,
    ) -> tuple[ChatResponseUsage | ImageResponseUsage | None, UsageCharge | None]:
        """Get dummy usage and charge data"""
        usage = None
        usage_charge = None

        if settings.ENABLE_USAGE_TRACKING or settings.ENABLE_COST_TRACKING:
            if ai_model_ep.ai_model.functional_type == AIModelFunctionalTypeEnum.TEXT_GENERATION:
                usage = ChatResponseUsage(total_tokens=1222, prompt_tokens=1000, completion_tokens=222)
            else:
                usage = ImageResponseUsage(number_of_images=1, model=ai_model_ep.ai_model.model_name, options={})

            if settings.ENABLE_COST_TRACKING:
                usage_charge = UsageCharge(cost=0.11, charge=0.22)

        return (usage, usage_charge)

    # -------------------------------------------------------------------------

    def _model_specific_message(
        self,
        ai_model_ep: AIModelEndpoint,
    ):
        return (
            f"This test message is from Model {ai_model_ep.ai_model.model_name}:"
            f"{ai_model_ep.ai_model.provider} with API Provider {ai_model_ep.api.provider}"
        )

    def _create_dummy_chunk(self, content: str, index: int = 0, finish: bool = False):
        """Create a dummy chunk for both sync and async streaming"""
        return type(
            "DummyChunk",
            (),
            {
                "id": "dummy-stream-id",
                "object": "chat.completion.chunk",
                "created": 1738776944,
                "system_fingerprint": "fp_dummy",
                "choices": [
                    type(
                        "Choice",
                        (),
                        {
                            "index": index,
                            "delta": type("Delta", (), {"content": content}),
                            "finish_reason": "stop" if finish else None,
                        },
                    )()
                ],
                "usage": None if not finish else {"total_tokens": 1000, "prompt_tokens": 333, "completion_tokens": 777},
            },
        )

    def _get_stream_text(self, ai_model_ep: AIModelEndpoint) -> str:
        """Get the text to be streamed"""
        base_text = self._model_specific_message(ai_model_ep)

        small_text = f"One\n Two\n Three\n Four\n Five\n Six Seven Eight Nine Ten"  # noqa: F541, F841
        large_text = """

Here is a detailed analysis of the topic:

1. First Important Point
   - Key insight A
   - Supporting detail B
   - Related concept C

2. Second Important Point
   - Technical aspect X
   - Implementation Y
   - Best practice Z

3. Final Considerations
   - Future implications
   - Recommendations
   - Next steps

Let me know if you need any clarification!"""

        markdown_text = """"
The issue with newlines being stripped in your code appears to be in how you're handling the incoming text deltas.
Looking at your logs, the newline characters (\n) are actually present in the `text_delta` values.

Here are a few potential solutions:

1. First, ensure your `onTextToken` callback preserves the newlines when displaying or processing the text. For example:

```javascript
onTextToken: (choice_index, content_index, text) => {
  // Preserve newlines by using pre-formatted elements if displaying in HTML
  const formattedText = text.replace(/\n/g, '<br>');
  // Or if appending to a textarea/console
  console.log(text); // Don't modify the text at all
},
```

2. If you're concatenating the tokens, make sure you're not accidentally stripping newlines:

```javascript
let fullText = '';
const processor = new StreamProcessor({
  onTextToken: (choice_index, content_index, text) => {
    fullText += text; // Don't modify the text, append as-is
    // If displaying in HTML
    element.innerHTML = fullText.replace(/\n/g, '<br>');
  }
});
```

3. You might also want to modify your `parseSSEEvent` method to explicitly preserve newlines:

```javascript
parseSSEEvent(eventString) {
  const lines = eventString.split(/\r?\n/);
  // ... rest of your parsing code ...

  if (data.length) {
    try {
      const dataString = data.join('\n'); // Preserve newlines when joining
      event.data = JSON.parse(dataString);
    } catch (e) {
      console.error('Error parsing JSON:', e);
      return null;
    }
  }

  return event;
}
```


Looking at your log output, there are tokens containing newlines,
so the issue is likely in how you're handling these tokens after receiving them rather than in the streaming process.

        """  # noqa: F841
        return f"{base_text}\n\n\n{large_text}"

    def _create_streaming_response(
        self, chunk_deltas: list[ChatResponseChoiceDelta]
    ) -> tuple[StreamingChatResponse | None, AIModelCallResponse | None]:
        """Process streaming chunks and create appropriate response"""
        stream_response = None

        if chunk_deltas:
            response_chunk = self.streaming_manager.update(choice_deltas=chunk_deltas)
            stream_response = StreamingChatResponse(
                id=None,
                data=response_chunk,
            )

        return stream_response

    def _process_chunk(self, chunk) -> StreamingChatResponse | None:
        """Process a single chunk for both sync and async streaming"""
        if chunk.usage:
            usage = ChatResponseUsage(
                total_tokens=chunk.usage["total_tokens"],
                prompt_tokens=chunk.usage["prompt_tokens"],
                completion_tokens=chunk.usage["completion_tokens"],
            )
            self.streaming_manager.update_usage(usage)

        if chunk.choices:
            choice_deltas = [
                ChatResponseChoiceDelta(
                    index=choice.index,
                    finish_reason=choice.finish_reason if hasattr(choice, "finish_reason") else None,
                    content_deltas=[
                        ChatResponseTextContentItemDelta(
                            index=0,
                            role="assistant",
                            text_delta=choice.delta.content or "",
                        )
                    ],
                    metadata={},
                )
                for choice in chunk.choices
            ]
            return self._create_streaming_response(choice_deltas)
        return None, None

    def sync_stream_generator(self, ai_model_ep: AIModelEndpoint):
        """Synchronous streaming generator"""
        import time

        text = self._get_stream_text(ai_model_ep)
        words = text.split(" ")

        for i, word in enumerate(words):
            time.sleep(0.04)  # Simulate network delay
            is_last = i == len(words) - 1
            yield self._create_dummy_chunk(word + " ", index=0, finish=is_last)

    async def async_stream_generator(self, ai_model_ep: AIModelEndpoint):
        """Asynchronous streaming generator"""
        text = self._get_stream_text(ai_model_ep)
        words = text.split(" ")

        for i, word in enumerate(words):
            await asyncio.sleep(0.04)  # Simulate network delay
            is_last = i == len(words) - 1
            yield self._create_dummy_chunk(word + " ", index=0, finish=is_last)

    def handle_streaming_response_sync(
        self,
        stream_generator,
        model_endpoint: AIModelEndpoint,
    ) -> Generator[tuple[StreamingChatResponse | SSEErrorResponse, AIModelCallResponse | None], None, None]:
        """Handle synchronous streaming response"""

        try:
            for chunk in stream_generator:
                stream_response = self._process_chunk(chunk)
                if stream_response:
                    yield stream_response, None

            # API has stopped streaming, send done-chunk and get final response
            done_chunk = self.streaming_manager.get_streaming_done_chunk()
            yield done_chunk, None

            # Final response after stream ends
            final_response = self.streaming_manager.get_final_response()
            yield None, final_response
            return

        except Exception as e:
            logger.exception(f"Error during streaming: {e}")
            yield (
                SSEErrorResponse(
                    data=SSEErrorData(
                        error_code=SSEErrorCode.external_api_error,
                        message="External API Server Error",
                        details={"error_type": type(e).__name__},
                    )
                ),
                None,
            )

    async def handle_streaming_response_async(
        self,
        stream_generator,
        model_endpoint: AIModelEndpoint,
    ) -> AsyncGenerator[tuple[StreamingChatResponse | SSEErrorResponse, AIModelCallResponse | None], None]:
        """Handle asynchronous streaming response"""

        try:
            async for chunk in stream_generator:
                stream_response = self._process_chunk(chunk)
                if stream_response:
                    yield stream_response, None

            # API has stopped streaming, send done-chunk and get final response
            done_chunk = self.streaming_manager.get_streaming_done_chunk()
            yield done_chunk, None

            # Final response after stream ends
            final_response = self.streaming_manager.get_final_response()
            yield None, final_response
            return

        except Exception as e:
            logger.exception(f"Error during streaming: {e}")
            yield (
                SSEErrorResponse(
                    data=SSEErrorData(
                        error_code=SSEErrorCode.external_api_error,
                        message="External API Server Error",
                        details={"error_type": type(e).__name__},
                    )
                ),
                None,
            )



"""
Settings and configuration for Dhenara-AI.
"""

import importlib
import logging
import os
from typing import Any
from dhenara.ai.config import default_settings

logger = logging.getLogger(__name__)


class Settings:
    """
    Settings class that loads default settings and user-defined settings.
    User settings can be defined in standard locations without needing environment variables.
    """

    def __init__(self):
        self._settings: dict[str, Any] = {}
        self._load_default_settings()
        self._load_user_settings()

    def _load_default_settings(self):
        for setting in dir(default_settings):
            if setting.isupper():
                self._settings[setting] = getattr(default_settings, setting)

    def _load_user_settings(self):
        """
        Search for user settings in standard locations and load them.
        Priority order:
        1. Current working directory
        2. User's home directory
        3. System-wide directory (optional)
        """
        possible_files = [
            os.path.join(os.getcwd(), "dhenara_config.py"),
            # os.path.expanduser("~/.dhenara_config.py"),
            # "/etc/dhenara_config.py",  # for system-wide settings
        ]

        for file_path in possible_files:
            if os.path.isfile(file_path):
                self._import_settings(file_path)
                logger.info(f"Loaded user settings from {file_path}")
                break  # Stop at the first found settings file

    def _import_settings(self, file_path: str):
        spec = importlib.util.spec_from_file_location("user_settings", file_path)
        if spec and spec.loader:
            user_settings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(user_settings)
            for setting in dir(user_settings):
                if setting.isupper():
                    self._settings[setting] = getattr(user_settings, setting)
        else:
            logger.error(f"Unable to load settings from {file_path}")

    def __getattr__(self, name: str) -> Any:
        """Get a setting value"""
        try:
            return self._settings[name]
        except KeyError:
            raise AttributeError(f"Setting '{name}' not found")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set a setting value"""
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self._settings[name] = value

    def update(self, **kwargs: Any) -> None:
        """Update settings with new values"""
        for key, value in kwargs.items():
            if key.isupper():
                self._settings[key] = value

    @property
    def configured_settings(self) -> list[str]:
        """Return list of all configured settings"""
        return sorted(self._settings.keys())


# Create a global settings object
settings = Settings()



# AI model API response
from ._content_items._base import *
from ._content_items._tool_call import *
from ._content_items._structured_output import *
from ._content_items._chat_items import *
from ._content_items._image_items import *

from ._metadata import *
from ._status import *

from ._chat import *
from ._image import *
from ._model_call import *



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    ChatModelCostData,
    ChatModelSettings,
    FoundationModel,
)

DeepseekR1 = FoundationModel(
    model_name="DeepSeek-R1",
    display_name="DeepSeek-R1",
    provider=AIModelProviderEnum.DEEPSEEK,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_context_window_tokens=128000,
        max_output_tokens=8000,
    ),
    valid_options={},
    metadata={
        "details": "DeepSeek-R1 model, optimized for reasoning.",
    },
    order=20,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.60,
        output_token_cost_per_million=2.20,
    ),
)

CHAT_MODELS = [DeepseekR1]



import logging
from typing import Any

from pydantic import Field, model_validator

from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    ChatResponseUsage,
    ImageResponseUsage,
    UsageCharge,
)
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


class ValidOptionValue(BaseModel):
    """
    Represents a valid option configuration for an AI model parameter.
    """

    allowed_values: list[Any] = Field(
        ...,
        description="List of allowed values for this option",
    )
    default_value: Any = Field(
        ...,
        description="Default value for this option",
    )
    cost_sensitive: bool = Field(
        ...,
        description="Will this option affect api-cost or not",
    )
    description: str | None = Field(
        default=None,
        description="Optional description of what this option controls",
    )
    display_only: bool = Field(
        default=False,
        description=(
            "Whether this option is only for display purpose."
            "When set, this won't be send in API calls. Useful for deriving additional options"
        ),
    )

    @model_validator(mode="after")
    def validate_default_in_allowed_values(self) -> "ValidOptionValue":
        """Ensures the default value is among allowed values."""
        if self.default_value not in self.allowed_values:
            raise ValueError(
                f"Default value {self.default_value} must be one of {self.allowed_values}",
            )
        return self


class BaseCostData(BaseModel):
    # NOTE: Default should be None to avoid wrong cost calculation
    # without proper overrides of standard foundation models in the package
    cost_multiplier_percentage: float | None = Field(
        default=None,
        description=(
            "Cost multiplication percentage f any."
            "Use this field to offset orgianl cost you paid to API provider with your additional expences/margin"
        ),
    )

    def calculate_usage_charge(self, usage) -> UsageCharge:
        raise NotImplementedError("calculate_usage_charge() not implemented")

    def get_charge(self, cost: float):
        if self.cost_multiplier_percentage:
            charge = round(
                cost * (1 + (self.cost_multiplier_percentage / 100)),
                6,
            )
        else:
            charge = None
        return UsageCharge(cost=cost, charge=charge)


class ChatModelCostData(BaseCostData):
    input_token_cost_per_million: float = Field(
        ...,
        description="",
    )
    output_token_cost_per_million: float = Field(
        ...,
        description="",
    )

    def calculate_usage_charge(
        self,
        usage: ChatResponseUsage,
    ) -> UsageCharge:
        try:
            input_per_token_cost = self.input_token_cost_per_million / 1000000
            output_per_token_cost = self.output_token_cost_per_million / 1000000

            cost = round(
                usage.prompt_tokens * input_per_token_cost + usage.completion_tokens * output_per_token_cost,
                6,
            )

            return self.get_charge(cost)
        except Exception as e:
            raise ValueError(f"calculate_usage_charge: Error: {e}")


class ImageModelCostData(BaseCostData):
    flat_cost_per_image: float | None = Field(
        default=None,
        description="Flat per image cost",
    )

    image_options_cost_data: list[dict] | None = Field(  # TODO: rename var
        default=None,
        description="Image options cost data",
    )

    @model_validator(mode="after")
    def _validate_cost_factores(self) -> "ImageModelCostData":
        if not (self.flat_cost_per_image or self.image_options_cost_data):
            raise ValueError("Either of flat_cost_per_image / image_options_cost_data must be set")
        if self.flat_cost_per_image and self.image_options_cost_data:
            raise ValueError("Set only one of flat_cost_per_image / image_options_cost_data is allowed")

        return self

    def calculate_usage_charge(
        self,
        usage: ImageResponseUsage,
    ) -> UsageCharge:
        try:
            cost_per_image = None
            if self.flat_cost_per_image:
                cost_per_image = self.flat_cost_per_image
            elif self.image_options_cost_data:
                cost_per_image = self.get_image_cost_with_options(
                    used_options=usage.options,
                )
            else:
                raise ValueError("calculate_image_charges: cost_per_image or cost options mapping is not set ")

            if cost_per_image is None:
                raise ValueError("calculate_image_charges: Failed to fix cost_per_image")

            cost = round(cost_per_image * usage.number_of_images, 6)
            return self.get_charge(cost)

        except Exception as e:
            raise ValueError(f"calculate_usage_charge: Error: {e}")

    # -------------------------------------------------------------------------
    def get_image_cost_with_options(self, used_options):
        if not self.image_options_cost_data:
            return None

        # Create a copy of used_options with standardized keys
        standardized_options = used_options.copy()

        # Remove keys that aren't used in cost data (like 'quality' )
        valid_keys = {key for data in self.image_options_cost_data for key in data.keys() if key != "cost_per_image"}
        standardized_options = {k: v for k, v in standardized_options.items() if k in valid_keys}

        for cost_data in self.image_options_cost_data:
            matches = True
            for key, value in standardized_options.items():
                if key not in cost_data or value not in cost_data[key]:
                    matches = False
                    break
            if matches:
                return cost_data["cost_per_image"]

        raise ValueError(
            f"get_image_cost_with_options: Failed to get price. "
            f"used_options={used_options}, image_options_cost_data={self.image_options_cost_data})"
        )


class ChatModelSettings(BaseModel):
    max_context_window_tokens: int | None = Field(
        default=None,
        description="Maximum context window size in tokens",
    )
    max_input_tokens: int | None = Field(
        default=None,
        description="Maximum input tokens allowed",
    )
    max_output_tokens: int = Field(
        ...,
        description="Maximum output tokens allowed",
    )
    # Reasoning settings
    supports_reasoning: bool = Field(
        default=False,
        description="If reasoning is supported or not",
    )
    max_reasoning_tokens: int | None = Field(
        default=None,
        description="Maximum reasoning tokens allowed",
    )
    max_output_tokens_reasoning_mode: int | None = Field(
        default=None,
        description="Maximum output tokens (including reasoning) in reasoning mode",
    )

    @model_validator(mode="after")
    def _set_token_limits(self) -> "ChatModelSettings":
        # Store original values
        input_tokens = self.max_input_tokens
        context_tokens = self.max_context_window_tokens
        output_tokens = self.max_output_tokens

        if not output_tokens:
            raise ValueError("set_token_limits: max_output_tokens must be specified")

        if not (input_tokens or context_tokens):
            raise ValueError("set_token_limits: max_input_tokens or max_context_window_tokens must be specified")

        # Calculate values without direct assignment
        values_to_update = {}
        if input_tokens is not None:
            values_to_update["max_context_window_tokens"] = input_tokens + output_tokens
        elif context_tokens is not None:
            values_to_update["max_input_tokens"] = context_tokens - output_tokens

        # fmt: off
        if self.supports_reasoning:
            if (
                self.max_reasoning_tokens
                and self.max_output_tokens_reasoning_mode
                and self.max_reasoning_tokens > self.max_output_tokens_reasoning_mode
            ):
                raise ValueError(
                    "set_token_limits: max_reasoning_tokens must be less than max_output_tokens_reasoning_mode"
                )
        # fmt: on

        # Update the model's dict directly
        for key, value in values_to_update.items():
            self.__dict__[key] = value

        return self


class ImageModelSettings(BaseModel):
    max_words: int | None = Field(
        None,
        description="Maximum word count, if applicable",
    )


class BaseAIModel(BaseModel):
    """
    Pydantic model representing an AI model configuration with options validation.
    """

    provider: AIModelProviderEnum = Field(
        ...,
        description="The AI model provider",
    )
    functional_type: AIModelFunctionalTypeEnum = Field(
        ...,
        description="Type of AI model functionality",
    )
    model_name: str = Field(
        ...,
        max_length=300,
        description="Model name used in API calls",
    )
    display_name: str = Field(
        ...,
        max_length=300,
        description="Display name for the model",
    )

    order: int = Field(
        0,
        description="Order for display purposes",
    )
    enabled: bool = Field(
        True,  # noqa: FBT003
        description="Whether the model is enabled",
    )
    beta: bool = Field(
        False,  # noqa: FBT003
        description="Whether the model is in beta",
    )

    settings: ChatModelSettings | ImageModelSettings | None = Field(
        default=None,
        description="Settings",
    )
    valid_options: dict[str, ValidOptionValue] | None = Field(
        default=None,
        description="Configured valid options and their allowed values",
    )
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional metadata if needed",
    )

    cost_data: ChatModelCostData | ImageModelCostData | None = Field(
        None,
        description="Optional Cost data",
    )

    reference_number: str | None = Field(
        None,
        description="Optional unique reference number",
    )

    @property
    def model_name_with_version_suffix(self):
        version_suffix = self.metadata.get("version_suffix", None)
        if version_suffix:
            return f"{self.model_name}{version_suffix}"
        else:
            return self.model_name

    @property
    def is_foundation_model(self) -> bool:
        return isinstance(self, FoundationModel)

    @model_validator(mode="after")
    def _validate_model_options(self) -> "BaseAIModel":
        """Validates model options against foundation model if present."""
        # Option validation is not done for foundation models
        if self.is_foundation_model:
            return self

        # If, not foundation model, ideally, options should be set, at least as an empty dict
        if self.foundation_model is None or self.valid_options is None:
            raise ValueError(
                "Either set the foundation model, or set the `valid_options` for this model.\
 If there are no options required for the models, set `valid_options` to an empty dict, but not as None"
            )

        # Validate that all options are present in foundation model
        invalid_options = set(self.valid_options.keys()) - set(
            self.foundation_model.valid_options.keys(),
        )
        if invalid_options:
            raise ValueError(
                f"Invalid options found: {invalid_options}. Must be subset of foundation model options.",
            )

        # Validate option values against foundation model
        for option_name, option_config in self.valid_options.items():
            foundation_config = self.foundation_model.valid_options[option_name]
            invalid_values = set(option_config.allowed_values) - set(
                foundation_config.allowed_values,
            )
            if invalid_values:
                raise ValueError(
                    f"Invalid values for option {option_name}: {invalid_values}",
                )

        return self

    @model_validator(mode="after")
    def _validate_settings(self) -> "BaseAIModel":
        if not self.settings:
            return self

        (setting_model, cost_model) = AIModel.get_pydantic_model_classes(self.functional_type)
        if not isinstance(self.settings, setting_model):
            raise ValueError(f"Settings should be instance of {setting_model} for {self.functional_type} models")
        if self.is_foundation_model and not isinstance(self.cost_data, cost_model):
            raise ValueError(f"For {self.functional_type} models, cost data must be set and of type {cost_model}")

        return self

    @model_validator(mode="after")
    def _validate_names(self) -> "BaseAIModel":
        if not self.display_name:
            self.display_name = self.model_name

        return self

    def validate_options(self, options: dict[str, Any]) -> bool:
        """
        Validates if the provided options conform to the model's valid options.

        Args:
            options: Dictionary of option name to value mappings

        Returns:
            bool: True if options are valid, False otherwise
        """
        try:
            self._validate_options_strict(options)
            return True
        except ValueError as e:
            logger.error(f"validate_options Fails: {e}")
            return False

    def _validate_options_strict(self, options: dict[str, Any]) -> None:
        """
        Strictly validates options and raises ValueError for invalid options.

        Args:
            options: Dictionary of option name to value mappings

        Raises:
            ValueError: If any option is invalid
        """
        _valid_options = self.get_valid_options()

        invalid_options = set(options.keys()) - set(_valid_options.keys())
        if invalid_options:
            raise ValueError(f"Unknown options: {invalid_options}")

        for option_name, value in options.items():
            valid_values = _valid_options[option_name].allowed_values
            if value not in valid_values:
                raise ValueError(
                    f"Invalid value for {option_name}: {value}. Must be one of {valid_values}",
                )

    def get_options_with_defaults(self, options: dict[str, Any]) -> dict[str, Any]:
        """
        Returns a complete options dictionary with defaults for missing values.

        Args:
            options: Partial options dictionary

        Returns:
            dict: Complete options dictionary with defaults
        """
        _valid_options = self.get_valid_options()

        self._validate_options_strict(options)

        complete_options = {}
        for option_name, option_config in _valid_options.items():
            complete_options[option_name] = options.get(
                option_name,
                option_config.default_value,
            )

        return complete_options

    def _get_attribute(self, attr_name: str):
        """Generic method to get attributes with foundation model fallback

        Args:
            attr_name: Name of the attribute to get
        Returns:
            Value of the attribute from this model or foundation model
        """
        if self.is_foundation_model:
            return getattr(self, attr_name)

        value = getattr(self, attr_name)
        if not value:
            return getattr(self.foundation_model, f"get_{attr_name}")()
        return value

    def get_settings(self):
        return self._get_attribute("settings")

    def _get_valid_options(self):
        return self._get_attribute("valid_options")

    def get_valid_options(self, include_display_only: bool = False) -> dict[str, ValidOptionValue]:
        """
        Get valid options excluding display-only options.

        Returns:
            dict: Dictionary of valid options without display-only options
        """
        options = self._get_valid_options()
        if include_display_only:
            return options

        if options:
            for key in list(options.keys()):
                if options[key].display_only:
                    del options[key]
        return options

    def get_cost_data(self):
        return self._get_attribute("cost_data")

    # -------------------------------------------------------------------------
    @classmethod
    def get_pydantic_model_classes(cls, functional_type):
        if functional_type == AIModelFunctionalTypeEnum.TEXT_GENERATION:
            setting_model = ChatModelSettings
            cost_model = ChatModelCostData
        elif functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
            setting_model = ImageModelSettings
            cost_model = ImageModelCostData
        else:
            raise ValueError(f"Functional type {functional_type} is not implemented ")

        return (setting_model, cost_model)


class FoundationModel(BaseAIModel):
    """Foundation model implementation with ability to create derived models."""

    def create_instance(
        self,
        model_name: str | None = None,
        display_name: str | None = None,
        order: int | None = None,
        enabled: bool = True,
        beta: bool = False,
        valid_options: dict[str, ValidOptionValue] | None = None,
        metadata: dict[str, Any] | None = None,
        reference_number: str | None = None,
        **kwargs,
    ) -> "AIModel":
        """
        Creates an AIModel instance based on this foundation model.

        Args:
            model_name: Optional model name (defaults to foundation model's name)
            display_name: Optional display name (defaults to model_name)
            order: Optional display order (defaults to foundation model's order)
            enabled: Whether the model is enabled (defaults to True)
            beta: Whether the model is in beta (defaults to False)
            valid_options: Optional valid options (defaults to foundation model's options)
            metadata: Optional metadata (will be merged with foundation model's metadata)
            reference_number: Optional reference number
            **kwargs: Additional model parameters

        Returns:
            AIModel: A new AI model instance based on this foundation model
        """

        # Use foundation model values as defaults
        model_name = model_name or self.model_name
        display_name = display_name or model_name
        order = order if order is not None else self.order
        valid_options = valid_options or self.valid_options
        merged_metadata = {**self.metadata, **(metadata or {})}

        # Create the model with foundation model as reference
        return AIModel(
            provider=self.provider,
            functional_type=self.functional_type,
            model_name=model_name,
            display_name=display_name,
            order=order,
            enabled=enabled,
            beta=beta,
            valid_options=valid_options,
            metadata=merged_metadata,
            foundation_model=self,
            reference_number=reference_number,
            **kwargs,
        )

    def clone(self, model_name: str) -> "AIModel":
        """Creates an exact clone of this foundation model as an AIModel instance."""
        return self.create_instance(
            model_name=model_name,
            display_name=self.display_name,
            order=self.order,
            enabled=self.enabled,
            beta=self.beta,
            valid_options=self.valid_options,
            metadata=self.metadata.copy(),
            reference_number=None,
        )


class AIModel(BaseAIModel):
    foundation_model: FoundationModel | None = Field(
        None,
        description="Matching foundation model for parameter preloading",
    )



import os
import re
from pathlib import Path

from setuptools import find_namespace_packages, setup

version = None
# Read version without importing the package
with open(os.path.join("src/dhenara/ai", "__init__.py")) as f:
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
    if version_match:
        version = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string")


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="dhenara-ai",
    version=version,
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src", include=["dhenara.*"]),
    install_requires=[
        "httpx>=0.28.0",
        "requests>=2.32.1",
        "asgiref>=3.8.0",
        "cryptography>=44.0.0",
        "aiohttp>=3.11.0",
        "pydantic>=2.10.0",
        "pyyaml>=6.0",
        "Pillow>=11.1.0",  # For images
        "openai>=1.65.0",
        "google-genai>=1.4.0",
        "anthropic>=0.49.0",
        # Cloud dependecies in extra
        "azure-ai-inference>=1.0.0b9",
        "boto3>=1.37.7",  # AWS
        "botocore>=1.37.7",  # AWS
    ],
    extras_require={
        "dev": [
            "pytest>=7.0.0",
            "pytest-asyncio",
            "pytest-cov",
            "black",
            "ruff",
            "add-trailing-comma",
        ],
        # TODO_FUTURE
        # "azure": [
        #    "azure-ai-inference>=1.0.0",
        # ],
        # "aws": [
        #    "boto3>=1.37.7",
        #    "botocore>=1.37.7",
        # ],
    },
    python_requires=">=3.10",
    description="Dhenara Package for Multi Provider AI-Model API calls",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dhenara",
    author_email="support@dhenara.com",
    url="https://github.com/dhenara/dhenara-ai",
    license="MIT",
    keywords="ai, llm, machine learning, language models",
    project_urls={
        "Homepage": "https://dhenara.com",
        "Documentation": "https://docs.dhenara.com/",
        "Bug Reports": "https://github.com/dhenara/dhenara-ai/issues",
        "Source Code": "https://github.com/dhenara/dhenara-ai",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
    ],
    include_package_data=True,
)



from typing import Any, Generic, TypeVar

from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel
from dhenara.ai.types.shared.platform import DhenaraAPIError


# -----------------------------------------------------------------------------
class ApiResponseStatus(BaseEnum):
    SUCCESS = "success"
    FAIL = "fail"
    ERROR = "error"
    PENDING = "pending"


# -----------------------------------------------------------------------------
# Add the missing status codes to ApiResponseMessageStatusCode
class ApiResponseMessageStatusCode(BaseEnum):
    SUCCESSFUL = "successful"
    PENDING = "pending"
    INVALID_INPUTS = "invalid_inputs"
    FAIL_ENDPOINT_ERROR = "failed_with_endpoint_error"
    FAIL_SERVER_ERROR = "failed_with_server_error"
    FAIL_FORBIDDEN = "failed_forbidden"
    FAIL_BAD_REQUEST = "failed_with_bad_request"
    NOT_AUTHENTICATED = "not_authenticated"
    NOT_AUTHORIZED = "not_authorized"
    PERMISSION_DENIED_GENERAL = "permission_denied"
    PERMISSION_DENIED_BY_WORKSPACE = "permission_denied_by_workspace"
    USAGE_EXCEEDED_LIMIT = "usage_exceeded_limit"
    # Auth-related status codes
    AUTH_MISSING_CREDENTIALS = "auth_missing_credentials"
    AUTH_INVALID_CREDENTIALS = "auth_invalid_credentials"
    AUTH_ACCOUNT_INACTIVE = "auth_account_inactive"
    AUTH_LOGIN_SUCCESS = "auth_login_success"
    AUTH_INVALID_REQUEST = "auth_invalid_request"
    AUTH_SERVER_ERROR = "auth_server_error"
    AUTH_MISSING_REFRESH_TOKEN = "auth_missing_refresh_token"
    AUTH_INVALID_REFRESH_TOKEN = "auth_invalid_refresh_token"
    AUTH_TOKEN_REFRESHED = "auth_token_refreshed"
    AUTH_LOGOUT_SUCCESS = "auth_logout_success"

    # Opps
    FAIL_NOT_FOUND = "fail_not_found"


# -----------------------------------------------------------------------------
class ApiResponseMessageType(BaseEnum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    SUCCESS = "success"


# -----------------------------------------------------------------------------
class ApiResponseMessage(BaseModel):
    type: ApiResponseMessageType
    status_code: ApiResponseMessageStatusCode
    message: str = ""
    message_data: dict | None = None
    show: bool = False  # Whether to show this message to user in case of FE

    def is_successful(self) -> bool:
        return self.status_code in [
            ApiResponseMessageStatusCode.SUCCESSFUL,
            ApiResponseMessageStatusCode.PENDING,
        ]


# -----------------------------------------------------------------------------
T = TypeVar("T", bound=BaseModel)


# -----------------------------------------------------------------------------
class ApiResponse(BaseModel, Generic[T]):
    status: ApiResponseStatus
    messages: list[ApiResponseMessage] = Field(default_factory=list)
    data: T | dict[str, Any] | None = None

    @property
    def is_success(self) -> bool:
        return self.status == ApiResponseStatus.SUCCESS

    @property
    def is_pending(self) -> bool:
        return self.status == ApiResponseStatus.PENDING

    @property
    def first_message(self) -> ApiResponseMessage | None:
        return self.messages[0] if self.messages else None

    def check_for_status_errors(self) -> None:
        """Raises an exception if the response indicates an error"""
        if not self.is_success:
            if self.first_message:
                return self.first_message
            return "Unknown error occurred"
        return None

    def raise_for_status(self) -> None:
        """Raises an exception if the response indicates an error"""
        error_msg = self.check_for_status_errors()
        if error_msg:
            raise DhenaraAPIError(
                message="Unknown error occurred",
                status_code=self.first_message.status_code or ApiResponseMessageStatusCode.FAIL_SERVER_ERROR,
                response={},
            )



# ruff: noqa: F401
from .fns import FoundationModelFns

from .constants import (
    ALL_FOUNDATION_MODELS,
    ALL_CHAT_MODELS,
    ALL_IMAGE_MODELS,
    # Provider specific
    ANTHROPIC_CHAT_MODELS,
    DEEPSEEK_CHAT_MODELS,
    GOOGLE_CHAT_MODELS,
    GOOGLE_IMAGE_MODELS,
    OPENAI_CHAT_MODELS,
    OPENAI_IMAGE_MODELS,
)



from enum import Enum
from typing import Any, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, alias_generators

T = TypeVar("T", bound="BaseModel")
# logger.debug(f"Pydantic version: {pydantic.__version__}")


class BaseEnum(str, Enum):
    """Base Enumeration class."""

    def __str__(self):
        return self.value

    @classmethod
    def values(cls) -> set[str]:
        """Get all values.

        Returns:
            set[str]: Set of all values
        """
        return {member.value for member in cls}


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(
        alias_generator=alias_generators.to_camel,
        populate_by_name=True,
        from_attributes=True,
        protected_namespaces=set(),
        # Enable detailed validation errors:
        validate_assignment=True,
        extra="forbid",
        arbitrary_types_allowed=True,
        # schema etra
        json_schema_extra={"examples": []},
        str_strip_whitespace=False,  # Don't set: Streaming responses will be terrible
        use_enum_values=True,
    )

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        dump_kwargs = {
            "exclude_unset": False,
            "by_alias": False,
            "exclude_none": True,
            "round_trip": False,
        }
        dump_kwargs.update(kwargs)

        return super().model_dump(**dump_kwargs)

    # def copy_with_changes(self: T, **changes) -> T:
    #    """Create a copy with specified changes."""
    #    data = self.model_dump()
    #    data.update(changes)
    #    return self.__class__.model_validate(data)

    # @classmethod
    # def safe_parse(cls: Type[T], data: dict[str, Any]) -> tuple[Optional[T], Optional[ValidationError]]:
    #    """Safely parse data without raising exceptions."""
    #    try:
    #        return cls.model_validate(data), None
    #    except ValidationError as e:
    #        return None, e

    # TODO_FUTURE: Comeup with better soution to mask selective fields
    def model_dump_safe(self, **kwargs):
        """Get a string representation with masked sensitive data"""
        # Get the regular model dump
        data = self.model_dump()
        # Create a copy to avoid modifying the original data
        return self._mask_sensitive_data(data.copy())

    def _mask_sensitive_data(self, data: dict) -> dict:
        """Recursively mask sensitive fields in a dictionary"""
        if not isinstance(data, dict):
            return data

        for key, value in data.items():
            if key == "api_key" and value:
                data[key] = f"{str(value)[:4]}...{str(value)[-4:]}" if len(str(value)) > 8 else value
            elif key == "credentials" and value and isinstance(value, dict):
                data[key] = {
                    k: f"{str(v)[:4]}...{str(v)[-4:]}" if v and len(str(v)) > 8 else v for k, v in value.items()
                }
            elif isinstance(value, dict):
                data[key] = self._mask_sensitive_data(value)
            elif isinstance(value, list):
                data[key] = [self._mask_sensitive_data(i) if isinstance(i, dict) else i for i in value]
        return data

    def _get_masked_representation(self) -> str:
        masked_data = self.model_dump_safe()

        # import json
        # return json.dumps(self.model_dump_safe(), indent=2)

        # Format it like the default __repr__ but with masked data
        fields = ", ".join(f"{k}={repr(v)}" for k, v in masked_data.items())  # noqa: RUF010
        return f"{self.__class__.__name__}({fields})"



from pydantic import Field, model_validator

from dhenara.ai.types.shared import BaseEnum, BaseModel


class ResourceConfigItemTypeEnum(BaseEnum):
    """Enumeration of available resource model types."""

    ai_model_endpoint = "ai_model_endpoint"
    rag_endpoint = "rag_endpoint"
    search_endpoint = "search_endpoint"


class ResourceQueryFieldsEnum(BaseEnum):
    """Enum defining all possible query fields for resources."""

    model_name = "model_name"
    model_display_name = "model_display_name"
    api_provider = "api_provider"
    reference_number = "reference_number"


class ResourceQueryMapping:
    """Static Class defining the mapping between resource types and their allowed query fields."""

    MAPPINGS: dict[ResourceConfigItemTypeEnum, list[ResourceQueryFieldsEnum]] = {
        ResourceConfigItemTypeEnum.ai_model_endpoint: [
            ResourceQueryFieldsEnum.model_name,
            ResourceQueryFieldsEnum.model_display_name,
            ResourceQueryFieldsEnum.api_provider,
        ],
    }

    @classmethod
    def get_allowed_fields(cls, resource_type: ResourceConfigItemTypeEnum) -> list[str]:
        """
        Get allowed query fields for a resource type.

        Args:
            resource_type: The type of resource

        Returns:
            List of allowed field names
        """
        return [field.value for field in cls.MAPPINGS.get(resource_type, [])]


class ResourceConfigItem(BaseModel):
    """
    ResourceConfigItem configuration model with mutually exclusive fields for object parameters
    or fetch query.

    Attributes:
        item_type: Type of the resource model
        query: Optional query string for fetching resource details
        is_default: Flag to mark default resource
    """

    item_type: ResourceConfigItemTypeEnum = Field(
        ...,
        description="Type of the resource item",
    )
    query: dict | None = Field(
        default=None,
        description="Query dict or list of query dicts for fetching resource details",
        examples=[
            {"model_name": "claude-sonet-3.5-v2"},
            {
                "model_name": "claude-sonet-3.5-v2",
                "api_provider": "anthropic",
            },
        ],
    )
    is_default: bool = Field(
        default=False,
        description="Is default resource or not. Only one default is allowed in a list of resources",
    )

    @model_validator(mode="after")
    def validate_exclusive_fields(self) -> "ResourceConfigItem":
        """Validates mutually exclusive fields and query structure."""
        # Validate query keys based on model type
        for key in self.query.keys():
            allowed_fields = ResourceQueryMapping.get_allowed_fields(self.item_type)
            if key not in allowed_fields:
                raise ValueError(f"Unsupported query key `{key}`")

        return self

    def is_same_as(self, other: "ResourceConfigItem") -> bool:
        """
        Compares two ResourceConfigItem objects for equality based on their type and identifiers.

        Args:
            other: Another ResourceConfigItem object to compare with

        Returns:
            bool: True if the resources represent the same entity, False otherwise
        """
        # First check if object types match
        if self.item_type != other.item_type:
            return False

        # Case 2: Both have query
        if self.query and other.query:
            return self.query == other.query

        # Case 3: Mixed cases are considered different
        return False

    # Factory methods
    @classmethod
    def with_model(cls, model_name: str) -> list["ResourceConfigItem"]:
        return [
            cls(
                item_type=ResourceConfigItemTypeEnum.ai_model_endpoint,
                query={ResourceQueryFieldsEnum.model_name: model_name},
                is_default=True,
            )
        ]

    @classmethod
    def with_models(cls, model_names: list[str]) -> list["ResourceConfigItem"]:
        return [
            cls(
                item_type=ResourceConfigItemTypeEnum.ai_model_endpoint,
                query={ResourceQueryFieldsEnum.model_name: model_name},
                is_default=True if index == 0 else False,
            )
            for index, model_name in enumerate(model_names)
        ]

    @classmethod
    def get_model_names(cls, resources: list["ResourceConfigItem"]) -> list[str]:
        """
        Extracts all model names from a list of ResourceConfigItem objects.
        This is the reverse operation of `with_models`.

        Args:
            resources: List of ResourceConfigItem objects

        Returns:
            list[str]: List of model names
        """
        return [
            resource.query[ResourceQueryFieldsEnum.model_name]
            for resource in resources
            if (
                resource.item_type == ResourceConfigItemTypeEnum.ai_model_endpoint
                and resource.query
                and ResourceQueryFieldsEnum.model_name in resource.query
            )
        ]

    @classmethod
    def has_model(cls, resources: list["ResourceConfigItem"], model_name: str) -> bool:
        """
        Check whether a given model name is present in the supplied
        list of ResourceConfigItem objects.

        Args:
            resources: List of ResourceConfigItem objects
            model_name: The model name to look for

        Returns:
            bool: True if the model name exists, False otherwise
        """
        return model_name in cls.get_model_names(resources)



"""
Default settings for Dhenara-AI.
All settings can be overridden by creating a settings file and setting DHENARA_SETTINGS_MODULE.
"""

#  ---- Request /Content Settings
"""
Convert all inputs (prompt/context/instructions/call_config params) to
provider specific formats before api-calls.
Disable only if you want to take care of different provider formats manually
"""
ENABLE_INPUT_FORMAT_CONVERSION = True

# ----  Response Content Settings

"""Enable Usage Tracking"""
ENABLE_USAGE_TRACKING = True

"""
Enable cost data creation in response.
When set,ENABLE_USAGE_TRACKING  will be always true
"""
ENABLE_COST_TRACKING = True

"""
 Consolidate streaming responses, and returns a consolidated response at the end
"""
ENABLE_STREAMING_CONSOLIDATION = True

# TODO:  Implement below settings
API_TIMEOUT = 30
MAX_RETRIES = 3

# Debug Settings
DEBUG = False
ENABLE_DETAILED_LOGGING = False



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    ChatModelCostData,
    ChatModelSettings,
    FoundationModel,
)

Gemini25Pro = FoundationModel(
    model_name="gemini-2.5-pro-preview-05-06",
    display_name="Gemini 2.5 Pro Preview",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=1048576,
        max_output_tokens=65535,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-2.5 Pro model",
        "display_order": 10,
    },
    order=82,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=1.25,
        output_token_cost_per_million=10.0,
    ),
)

Gemini25Flash = FoundationModel(
    model_name="gemini-2.5-flash-preview-04-17",
    display_name="Gemini 2.5 Flash Preview",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=1048576,
        max_output_tokens=65535,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-2.5-flash model",
        "display_order": 10,
    },
    order=82,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.15,
        output_token_cost_per_million=3.50,
    ),
)


Gemini20Flash = FoundationModel(
    model_name="gemini-2.0-flash",
    display_name="Gemini 2 Flash",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=1048576,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-2.0-flash model",
        "display_order": 10,
    },
    order=82,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.10,
        output_token_cost_per_million=0.40,
    ),
)

Gemini20FlashLite = FoundationModel(
    model_name="gemini-2.0-flash-lite",
    display_name="Gemini 2 Flash Lite",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=1048576,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-2.0-flash-light model",
        "display_order": 10,
    },
    order=83,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.075,
        output_token_cost_per_million=0.30,
    ),
)

Gemini15Pro = FoundationModel(
    model_name="gemini-1.5-pro",
    display_name="Gemini 1.5 Pro",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=2097152,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-1.5-pro model, Optimized for complex reasoning tasks",
        "display_order": 91,
    },
    order=21,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=2.50,
        output_token_cost_per_million=10.0,
    ),
)
Gemini15Flash = FoundationModel(
    model_name="gemini-1.5-flash",
    display_name="Gemini 1.5 Flash",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.TEXT_GENERATION,
    settings=ChatModelSettings(
        max_input_tokens=1048576,
        max_output_tokens=8192,
    ),
    valid_options={},
    metadata={
        "details": "GoogleAI gemini-1.5-flash model",
        "display_order": 92,
    },
    order=20,
    cost_data=ChatModelCostData(
        input_token_cost_per_million=0.15,
        output_token_cost_per_million=0.60,
    ),
)

CHAT_MODELS = [
    Gemini20Flash,
    Gemini20FlashLite,
    Gemini15Flash,
    Gemini15Pro,
]



from dhenara.ai.types.genai.ai_model import (
    AIModelFunctionalTypeEnum,
    AIModelProviderEnum,
    FoundationModel,
    ImageModelCostData,
    ImageModelSettings,
    ValidOptionValue,
)

_imagen_valid_options = {
    "aspect_ratio": ValidOptionValue(
        allowed_values=["1:1", "9:16", "16:9", "4:3", "3:4"],
        default_value="1:1",
        cost_sensitive=True,
        description="Aspect ratio",
    ),
    "number_of_images": ValidOptionValue(
        allowed_values=list(range(1, 4)),
        default_value=1,
        cost_sensitive=True,
        description="Number of images",
    ),
    "add_watermark": ValidOptionValue(
        allowed_values=[True, False],
        default_value=False,
        cost_sensitive=False,
        description="Add watermark or not",
    ),
    "safety_filter_level": ValidOptionValue(
        allowed_values=[None, "block_low_and_above", "block_medium_and_above", "block_only_high"],
        default_value="block_only_high",
        cost_sensitive=False,
        description="Safety Filter Level",
    ),
    "person_generation": ValidOptionValue(
        allowed_values=[None, "dont_allow", "allow_adult", "allow_all"],
        default_value="allow_all",
        cost_sensitive=False,
        description="Person generation settings",
    ),
}


Imagen3 = FoundationModel(
    model_name="imagen-3.0-generate",
    display_name="imagen-3.0",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.IMAGE_GENERATION,
    settings=ImageModelSettings(),
    valid_options=_imagen_valid_options,
    metadata={
        "details": "Imagen 3.0",
        "version_suffix": "-002",  # NOTE: This is required for google models
    },
    order=1,
    cost_data=ImageModelCostData(
        flat_cost_per_image=0.04,
        image_options_cost_data=None,
    ),
)


Imagen3Fast = FoundationModel(
    model_name="imagen-3.0-fast-generate",
    display_name="imagen-3.0 Fast",
    provider=AIModelProviderEnum.GOOGLE_AI,
    functional_type=AIModelFunctionalTypeEnum.IMAGE_GENERATION,
    settings=ImageModelSettings(),
    valid_options=_imagen_valid_options,
    metadata={
        "details": "Imagen 3.0",
        "version_suffix": "-001",  # NOTE: This is required for google models
    },
    order=2,
    cost_data=ImageModelCostData(
        flat_cost_per_image=0.02,
        image_options_cost_data=None,
    ),
)
IMAGE_MODELS = [Imagen3, Imagen3Fast]



from typing import Generic, TypeVar

# -----------------------------------------------------------------------------
from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel


# -----------------------------------------------------------------------------
class ApiRequestActionTypeEnum(BaseEnum):
    """Common action types for endpoint operations."""

    # GET
    list = "list"
    get = "get"
    # POST
    create = "create"
    update = "update"
    delete = "delete"
    activate = "activate"
    deactivate = "deactivate"
    # Run
    run = "run"


# -----------------------------------------------------------------------------
T = TypeVar("T", bound=BaseModel)


class ApiRequest(BaseModel, Generic[T]):
    """
    Base arequest model for generic operations.

    Args:
        data: The actual data for the action
        action: The type of action to perform
    """

    data: T
    action: ApiRequestActionTypeEnum | None = Field(
        ...,
        description="Action to perform on the endpoint",
    )
    custom_action: str | None = Field(
        default=None,
        description="Custom Action to perform on the endpoint",
    )



import collections


def generic_obj_to_dict(obj):
    """
    Recursively convert an object to a dictionary.
    Handles lists, dictionaries, mappingproxy, protobuf messages, and objects with __dict__ attribute.
    """
    if isinstance(obj, dict):
        # If obj is a dictionary, apply the function to each value
        return {key: generic_obj_to_dict(value) for key, value in obj.items()}
    elif isinstance(obj, collections.abc.Mapping):
        # If obj is a mapping (including mappingproxy), convert to a dict
        return {key: generic_obj_to_dict(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        # If obj is a list, apply the function to each element
        return [generic_obj_to_dict(element) for element in obj]
    elif hasattr(obj, "__dict__"):
        # If obj is an object with a dictionary attribute, convert it
        return generic_obj_to_dict(obj.__dict__)
    else:
        # If obj is a basic type (int, str, etc.), return it as is
        return obj



# -----------------------------------------------------------------------------
from dhenara.ai.types.shared.base import BaseEnum


class PromptMessageRoleEnum(BaseEnum):
    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"



from typing import Any

from pydantic import Field, model_validator

from dhenara.ai.types.shared.base import BaseModel


class TextTemplateVariableProps(BaseModel):
    default: Any | None = Field(default=None, description="Default value for the parameter")
    allowed: list[Any] | None = Field(default=None, description="Allowed values")
    # type: str = Field(..., description="Type of the parameter (string, number, boolean, etc.)")
    # description: str | None = Field(default=None, description="Description of the parameter")
    # required: bool = Field(default=False, description="Whether the parameter is required")


class TextTemplate(BaseModel):
    """
    Enhanced template configuration for AI interactions.
    Supports two template syntaxes:

    1. Variable substitution with $var{variable_name}
    2. Expression-based templates with $expr{expressions}

    Regular braces {} are left untouched and treated as literal text.

    Escape sequences:
    - Use $$var{} to output a literal "$var{}" string
    - Use $$expr{} to output a literal "$expr{}" string

    NOTE: Parsing templates using this syntaxes are NOT handled within this (dhenara.ai) package.
    You need to use a separate template engine for that.
    """

    text: str = Field(
        description="Text template with $var{variables} and $expr{expressions}",
    )
    variables: dict[str, TextTemplateVariableProps | None] = Field(
        default_factory=dict,
        description="Variables/parameters for the template",
    )
    disable_checks: bool = Field(
        default=False,
    )

    @model_validator(mode="after")
    def validate_variables(self) -> "TextTemplate":
        """
        Validate that all defined variables appear in the template.
        Looks for $var{variable_name} patterns in the template text.
        Supports variables with dot notation (nested variables).
        """
        if not self.disable_checks:
            import re

            # Get variable names from the template using $var{} pattern
            text = self.text
            defined_vars = list(self.variables.keys())

            # Find all $var{...} patterns, but ignore escaped ones ($$var{...})
            # First replace all escaped patterns with a placeholder
            escaped_patterns = {}

            def replace_escaped(match):
                placeholder = f"__ESCAPED_{len(escaped_patterns)}__"
                escaped_patterns[placeholder] = match.group(0)
                return placeholder

            # Replace escaped patterns
            text_for_validation = re.sub(r"\$\$var{[^}]*}", replace_escaped, text)
            text_for_validation = re.sub(r"\$\$expr{[^}]*}", replace_escaped, text_for_validation)

            # Now find all non-escaped variable patterns
            var_pattern = re.compile(r"\$var{([^}]+)}")
            used_vars_full = [match.group(1).strip() for match in var_pattern.finditer(text_for_validation)]

            # Extract root variable names from possibly nested paths (e.g., 'task_spec.description' -> 'task_spec')
            used_vars = set()
            for var_path in used_vars_full:
                root_var = var_path.split(".")[0].strip()
                used_vars.add(root_var)

            # Check for variables defined but not used in template
            missing_in_text = [var for var in defined_vars if var not in used_vars]
            if missing_in_text:
                raise ValueError(f"Variables {missing_in_text} are defined but not used in the template text")

        return self

    def get_variable_names(self) -> dict[str, Any]:
        """Get a dictionary of variable default values."""
        return list(self.variables.keys())

    def get_args_default_values(self) -> dict[str, Any]:
        """Get a dictionary of variable default values."""
        return {key: props.default for key, props in self.variables.items() if props and props.default is not None}

    def format(self, **kwargs) -> str:
        """
        Returns the template text without processing any variables or expressions.
        Variable substitution and expression evaluation are NOT handled within this package.

        This is NOT a formatting method - it just returns the raw template.
        """
        # Just return the raw template text - no Python formatting
        return self.text


class ObjectTemplate(BaseModel):
    """
    Template configuration for retrieving objects from expressions.
    Unlike TextTemplate, this preserves the type of the evaluated expression
    rather than converting to string.

    Uses the $expr{} syntax for expressions.
    """

    expression: str = Field(
        description="Expression template containing a $expr{expression} that returns an object",
    )



import logging

from anthropic.types import (
    ContentBlock,
    Message,
    MessageStreamEvent,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    RedactedThinkingBlock,
    SignatureDelta,
    TextBlock,
    TextDelta,
    ThinkingBlock,
    ThinkingDelta,
    ToolUseBlock,
)

from dhenara.ai.providers.anthropic import AnthropicClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseContentItem,
    ChatResponseContentItemDelta,
    ChatResponseReasoningContentItem,
    ChatResponseReasoningContentItemDelta,
    ChatResponseStructuredOutput,
    ChatResponseStructuredOutputContentItem,
    ChatResponseTextContentItem,
    ChatResponseTextContentItemDelta,
    ChatResponseToolCallContentItem,
    ChatResponseUsage,
    StreamingChatResponse,
)
from dhenara.ai.types.genai.dhenara import ChatResponseToolCall
from dhenara.ai.types.shared.api import SSEErrorResponse

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class AnthropicChat(AnthropicClientBase):
    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: dict | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("inputs must be validated with validate_inputs() before api calls")

        messages = []
        user = self.config.get_user()

        # Process system instructions
        system_prompt = None
        if instructions:
            if not (isinstance(instructions, dict) and "content" in instructions.keys()):
                raise ValueError(
                    f"Invalid Instructions format. "
                    f"Instructions should be processed and passed in prompt format. Value is {instructions} "
                )
            system_prompt = instructions["content"]  # Extract text from system prompt

        # Add previous messages and current prompt
        if context:
            messages.extend(context)

        messages.append(prompt)

        # Prepare API call arguments
        chat_args = {
            "model": self.model_name_in_api_calls,
            "messages": messages,
            "stream": self.config.streaming,
        }

        if system_prompt:
            chat_args["system"] = system_prompt

        if user:
            chat_args["metadata"] = {"user_id": user}

        max_output_tokens, max_reasoning_tokens = self.config.get_max_output_tokens(self.model_endpoint.ai_model)
        if max_output_tokens is not None:
            chat_args["max_tokens"] = max_output_tokens

        if max_reasoning_tokens is not None:
            chat_args["thinking"] = {
                "type": "enabled",
                "budget_tokens": max_reasoning_tokens,
            }

        if self.config.options:
            chat_args.update(self.config.options)

        # ---  Tools ---
        if self.config.tools:
            chat_args["tools"] = self.formatter.format_tools(
                tools=self.config.tools,
                model_endpoint=self.model_endpoint,
            )

        if self.config.tool_choice:
            chat_args["tool_choice"] = self.formatter.format_tool_choice(
                tool_choice=self.config.tool_choice,
                model_endpoint=self.model_endpoint,
            )

        # --- Structured Output ---
        # Anthropic uses the tool system for structured output
        if self.config.structured_output:
            # For Anthropic, we need to set up tool calling
            if "tools" not in chat_args:
                chat_args["tools"] = []

            # Add structured output as a tool
            structured_tool = self.formatter.format_structured_output(
                structured_output=self.config.structured_output,
                model_endpoint=self.model_endpoint,
            )
            chat_args["tools"].append(structured_tool)

            # Enforce this tool
            if max_reasoning_tokens is not None:
                # TODO_FUTURE: Revisit this later if API improves in future
                # Currently when enforced tool use in thiking mode,  API flags error as
                # 'Thinking may not be enabled when tool_choice forces tool use.'
                # The irony is that they don't have a structured-output mode either
                chat_args["tool_choice"] = {"type": "auto"}
            else:
                chat_args["tool_choice"] = {
                    "type": "tool",
                    "name": structured_tool["name"],
                }

        return {"chat_args": chat_args}

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        response = self._client.messages.create(**chat_args)
        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        response = await self._client.messages.create(**chat_args)
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        stream = self._client.messages.create(**chat_args)
        return stream

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat_args = api_call_params["chat_args"]
        stream = await self._client.messages.create(**chat_args)
        return stream

    def parse_stream_chunk(
        self,
        chunk: MessageStreamEvent,
    ) -> StreamingChatResponse | SSEErrorResponse | None:
        """Handle streaming response with progress tracking and final response"""

        processed_chunks = []

        # self.streaming_manager.message_metadata  is used to preserve params of initial message across chunks
        if isinstance(chunk, RawMessageStartEvent):
            message = chunk.message

            # Initialize message metadata
            self.streaming_manager.message_metadata = {
                "id": message.id,
                "model": message.model,
                "role": message.role,
                "type": type,
                "index": 0,  # Only one choice from Antropic
            }

            # Anthropic has a wieded way of reporint usage on streaming
            # On message_start, usage will have input tokens and few output tokens
            _usage = chunk.message.usage
            if _usage:
                # Initialize usage in self.streaming_manager
                usage = ChatResponseUsage(
                    total_tokens=0,
                    prompt_tokens=_usage.input_tokens,
                    completion_tokens=_usage.output_tokens,
                )
                self.streaming_manager.update_usage(usage)

        elif isinstance(chunk, RawContentBlockStartEvent):
            block_type = chunk.content_block.type
            if block_type == "redacted_thinking":
                content_deltas = [
                    ChatResponseReasoningContentItem(
                        index=chunk.index,
                        role=self.streaming_manager.message_metadata["role"],
                        metadata={
                            "redacted_thinking_data": chunk.content_block.data,
                        },
                    )
                ]

                choice_deltas = [
                    ChatResponseChoiceDelta(
                        index=self.streaming_manager.message_metadata["index"],
                        content_deltas=content_deltas,
                        metadata={},
                    )
                ]

                response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
                stream_response = StreamingChatResponse(
                    id=self.streaming_manager.message_metadata["id"],
                    data=response_chunk,
                )

                processed_chunks.append(stream_response)
            elif block_type in ["text", "thinking"]:
                pass
            else:
                logger.debug(f"anthropic: Unhandled content_block_type {block_type}")

        elif isinstance(chunk, RawContentBlockDeltaEvent):
            content_deltas = [
                self.process_content_item_delta(
                    index=chunk.index,
                    role=self.streaming_manager.message_metadata["role"],
                    delta=chunk.delta,
                )
            ]

            choice_deltas = [
                ChatResponseChoiceDelta(
                    index=self.streaming_manager.message_metadata["index"],
                    content_deltas=content_deltas,
                    metadata={},
                )
            ]
            response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
            stream_response = StreamingChatResponse(
                id=self.streaming_manager.message_metadata["id"],
                data=response_chunk,
            )
            processed_chunks.append(stream_response)
        elif isinstance(chunk, RawContentBlockStopEvent):
            pass
        elif isinstance(chunk, RawMessageDeltaEvent):
            # Update output tokens
            self.streaming_manager.usage.completion_tokens += chunk.usage.output_tokens
            self.streaming_manager.usage.total_tokens = (
                self.streaming_manager.usage.prompt_tokens + self.streaming_manager.usage.completion_tokens
            )

            # Update choice metatdata
            choice_deltas = [
                ChatResponseChoiceDelta(
                    index=self.streaming_manager.message_metadata["index"],
                    finish_reason=chunk.delta.stop_reason,
                    stop_sequence=chunk.delta.stop_sequence,
                    content_deltas=[],
                    metadata={},
                )
            ]
            response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
            stream_response = StreamingChatResponse(
                id=self.streaming_manager.message_metadata["id"],
                data=response_chunk,
            )
            processed_chunks.append(stream_response)

        elif isinstance(chunk, RawMessageStopEvent):
            pass
        else:
            logger.debug(f"anthropic: Unhandled message type {chunk.type}")

        return processed_chunks

    # API has stopped streaming, get final response

    def _get_usage_from_provider_response(
        self,
        response: Message,
    ) -> ChatResponseUsage:
        return ChatResponseUsage(
            total_tokens=response.usage.input_tokens + response.usage.output_tokens,
            prompt_tokens=response.usage.input_tokens,
            completion_tokens=response.usage.output_tokens,
        )

    def parse_response(self, response: Message) -> ChatResponse:
        usage, usage_charge = self.get_usage_and_charge(response)

        return ChatResponse(
            model=response.model,
            provider=self.model_endpoint.ai_model.provider,
            api_provider=self.model_endpoint.api.provider,
            usage=usage,
            usage_charge=usage_charge,
            choices=[
                ChatResponseChoice(
                    index=0,  # Only one choice
                    finish_reason=response.stop_reason,
                    stop_sequence=response.stop_sequence,
                    contents=[
                        self.process_content_item(
                            index=content_index,  # enumerate as Anthropic APIs doesn't provide index for non-streaming
                            role=response.role,
                            content_item=content_item,
                        )
                        for content_index, content_item in enumerate(response.content)
                    ],
                    metadata={},  # Choice metadata
                ),
            ],
            # Response Metadata
            metadata=AIModelCallResponseMetaData(
                streaming=False,
                duration_seconds=0,  # TODO
                provider_metadata={
                    "id": response.id,
                },
            ),
        )

    def process_content_item(
        self,
        index: int,
        role: str,
        content_item: ContentBlock,
    ) -> ChatResponseContentItem:
        if isinstance(content_item, TextBlock):
            return ChatResponseTextContentItem(
                index=index,
                role=role,
                text=content_item.text,
            )
        elif isinstance(content_item, ThinkingBlock):
            return ChatResponseReasoningContentItem(
                index=index,
                role=role,
                thinking_text=content_item.thinking,
                metadata={
                    "signature": content_item.signature,
                },
            )

        elif isinstance(content_item, RedactedThinkingBlock):
            return ChatResponseReasoningContentItem(
                index=index,
                role=role,
                metadata={
                    "redacted_thinking_data": content_item.data,
                },
            )

        elif isinstance(content_item, ToolUseBlock):
            raw_response = content_item.model_dump()
            try:
                tool_call = ChatResponseToolCall.from_anthropic_format(raw_response)
            except Exception as e:
                logger.exception(f"Error parsing tool call: {e}")
                tool_call = None

            # For anthropic, structed output reqs are send as tool_call
            if self.config.structured_output is not None:
                structured_output = ChatResponseStructuredOutput.from_tool_call(
                    raw_response=raw_response,
                    tool_call=tool_call,
                    config=self.config.structured_output,
                )
                return ChatResponseStructuredOutputContentItem(
                    index=index,
                    role=role,
                    structured_output=structured_output,
                )
            else:
                if tool_call:
                    return ChatResponseToolCallContentItem(
                        index=index,
                        role=role,
                        tool_call=tool_call,
                        metadata={},
                    )
                else:
                    return self.get_unknown_content_type_item(
                        index=index,
                        role=role,
                        unknown_item=content_item,
                        streaming=False,
                    )

        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=content_item,
                streaming=False,
            )

    def process_content_item_delta(
        self,
        index: int,
        role: str,
        delta,
    ) -> ChatResponseContentItemDelta:
        if isinstance(delta, TextDelta):
            return ChatResponseTextContentItemDelta(
                index=index,
                role=role,
                text_delta=delta.text,
            )
        elif isinstance(delta, ThinkingDelta):
            return ChatResponseReasoningContentItemDelta(
                index=index,
                role=role,
                thinking_text_delta=delta.thinking,
                metadata={},
            )
        elif isinstance(delta, SignatureDelta):
            return ChatResponseReasoningContentItemDelta(
                index=index,
                role=role,
                thinking_text_delta="",
                metadata={
                    "signature": delta.signature,
                },
            )
        # TODO: Tools Not supported in streaming yet
        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=delta,
                streaming=True,
            )
